import time
import numpy as np

from random import choices, seed, randint
from math import sqrt
from copy import deepcopy
from pydantic import BaseModel, validate_call
from typing import Union, List, Optional, Tuple, Dict
from itertools import product
from tqdm.contrib import tenumerate

import invertedai as iai
from invertedai.large.common import Region, REGION_MAX_SIZE
from invertedai.api.initialize import InitializeResponse
from invertedai.utils import get_default_agent_properties
from invertedai.error import InvertedAIError
from invertedai.common import (
    AgentProperties, 
    AgentState, 
    AgentType,
    Point,
    TrafficLightStatesDict   
)

AGENT_SCOPE_FOV_BUFFER = 60
ATTEMPT_PER_NUM_REGIONS = 15


@validate_call
def get_regions_default(
    location: str,
    total_num_agents: Optional[int] = None,
    agent_count_dict: Optional[Dict[AgentType,int]] = None,
    area_shape: Optional[Tuple[float,float]] = None,
    map_center: Optional[Tuple[float,float]] = (0.0,0.0),
    random_seed: Optional[int] = None, 
    display_progress_bar: Optional[bool] = True
) -> List[Region]:
    """
    A utility function to create a set of Regions to be passed into :func:`large_initialize` in
    a single convenient entry point. 

    Arguments
    ----------
    location:
        Location name in IAI format.

    total_num_agents:
        Deprecated. The total number of agents to initialize across all regions.

    agent_count_dict:
        The number of agents to place within the regions per specified agent type.

    area_shape:
        Contains the [width, height] to either side of the center of the rectangular area to be broken into 
        smaller regions (i.e. half of full width and height of the region). If this argument is not provided, 
        a bounding box around the location polygon from :func:`location_info` will be used.

    map_center:
        The coordinates of the center of the rectangular area to be broken into smaller regions. If
        this argument is not provided, default coordinates of (0,0) are used.

    random_seed:
        Controls the stochastic aspects of assigning agents to regions for reproducibility. If this
        argument is not provided, a random seed is chosen.

    display_progress_bar:
        A flag to control whether a command line progress bar is displayed for convenience.
    
    """
    
    if agent_count_dict is None:
        if total_num_agents is None:
            raise InvertedAIError(message=f"Must specify a number of agents within the regions.")
        else:
            agent_count_dict = {AgentType.car: total_num_agents}

    if area_shape is None:
        location_response = iai.location_info(
            location = location,
            rendering_center = map_center
        )
        polygon_x, polygon_y = [], []
        for pt in location_response.bounding_polygon:
            polygon_x.append(pt.x)
            polygon_y.append(pt.y)
        width = max(polygon_x) - min(polygon_x)
        height = max(polygon_y) - min(polygon_y)

        area_shape = (width/2,height/2)

    regions = iai.get_regions_in_grid(
        width = area_shape[0], 
        height = area_shape[1],
        map_center = map_center
    )

    new_regions = iai.get_number_of_agents_per_region_by_drivable_area(
        location = location,
        regions = regions,
        agent_count_dict = agent_count_dict,
        random_seed = random_seed,
        display_progress_bar = display_progress_bar
    )

    return new_regions


@validate_call
def get_regions_in_grid(
    width: float,
    height: float,
    map_center: Optional[Tuple[float,float]] = (0.0,0.0), 
    stride: Optional[float] = 50.0
) -> List[Region]:
    """
    A utility function to help initialize an area larger than 100x100m. This function breaks up an 
    area into a grid of 100x100m regions, each of which can be initialized directly. Each of these
    regions will only have its shaped specified and no agent information. 

    Arguments
    ----------
    width:
        Half of the horizontal size of the area which will be broken into smaller regions.

    height:
        Half of the vertical size of the area which will be broken into smaller regions.

    map_center:
        The coordinates of the center of the rectangular area to be broken into smaller regions. If
        this argument is not provided, default coordinates of (0,0) are used.

    stride:
        How far apart the centers of the 100x100m regions should be. Some overlap is recommended for 
        best results and if no argument is provided, a value of 50 is used.
    """

    def check_valid_center(center):
        return (map_center[0] - width) < center[0] < (map_center[0] + width) and \
            (map_center[1] - height) < center[1] < (map_center[1] + height)

    def get_neighbors(center):
        return [
            (center[0] + (i * stride), center[1] + (j * stride))
            for i, j in list(product(*[(-1, 1),]* 2))
        ]

    queue, centers = [map_center], []

    while queue:
        center = queue.pop(0)
        neighbors = filter(check_valid_center, get_neighbors(center))
        queue.extend([
            neighbor
            for neighbor in neighbors
            if neighbor not in queue and neighbor not in centers
        ])
        if center not in centers and check_valid_center(center):
            centers.append(center)
    
    regions = [None for _ in range(len(centers))]
    for i, center in enumerate(centers):
        regions[i] = Region.create_square_region(center=Point.fromlist(list(center)))

    return regions


@validate_call
def get_number_of_agents_per_region_by_drivable_area(
    location: str,
    regions: List[Region],
    total_num_agents: Optional[int] = None,
    agent_count_dict: Optional[Dict[AgentType,int]] = None,
    random_seed: Optional[int] = None,
    display_progress_bar: Optional[bool] = True
) -> List[Region]:
    """
    Takes a list of regions, calculates the driveable area for each of them using output from
    :func:`location_info`, then creates a new Region object with copied location and shape data and 
    inserts a number of car agents to be **sampled** into it proportional to its drivable surface 
    area relative to the other regions. Regions with no or a relatively small amount of drivable 
    surface will be removed.

    Arguments
    ----------
    location:
        Location name in IAI format.

    regions:
        A list of Regions which may or may not contain pre-existing agents. 

    total_num_agents:
        Deprecated. The total number of agents to initialize across all regions.

    agent_count_dict:
        The number of agents to place within the regions per specified agent type.

    random_seed:
        Controls the stochastic aspects of assigning agents to regions for reproducibility. If this
        argument is not provided, a random seed is chosen.

    display_progress_bar:
        A flag to control whether a command line progress bar is displayed for convenience.
    """

    if agent_count_dict is None:
        if total_num_agents is None:
            raise InvertedAIError(message=f"Must specify a number of agents within the regions.")
        else:
            agent_count_dict = {AgentType.car: total_num_agents}

    agent_list_types = []
    for agent_type, num_agents in agent_count_dict.items():
        agent_list_types = agent_list_types + [agent_type]*num_agents

    new_regions = [Region.copy(region) for region in regions]
    region_road_area = []
    total_drivable_area_ratio = 0

    if random_seed is not None:
        seed(random_seed)
    
    if display_progress_bar:
        iterable_regions = tenumerate(
            new_regions, 
            total=len(new_regions),
            desc=f"Calculating drivable surface areas"
        )
    else:
        iterable_regions = enumerate(new_regions)

    for i, region in iterable_regions:
        center_tuple = (region.center.x, region.center.y)
        birdview = iai.location_info(
            location=location,
            rendering_fov=int(region.size),
            rendering_center=center_tuple
        ).birdview_image.decode()

        birdview_arr_shape = birdview.shape
        total_num_pixels = birdview_arr_shape[0]*birdview_arr_shape[1]
        number_of_black_pix = np.sum(birdview.sum(axis=-1) == 0)

        drivable_area_ratio = (total_num_pixels-number_of_black_pix)/total_num_pixels 
        total_drivable_area_ratio += drivable_area_ratio
        region_road_area.append(drivable_area_ratio)

    # Select region in which to assign agents using drivable area as weight
    all_region_weights = [0]*len(new_regions)
    for i, drivable_ratio in enumerate(region_road_area):
        all_region_weights[i] = drivable_ratio/total_drivable_area_ratio
    random_indexes = choices(list(range(len(new_regions))), weights=all_region_weights, k=len(agent_list_types))

    for agent_id, ind in enumerate(random_indexes):
        new_regions[ind].agent_properties = new_regions[ind].agent_properties + get_default_agent_properties({agent_list_types[agent_id]:1})

    filtered_regions = []
    for region in new_regions:
        if len(region.agent_properties) > 0:
            filtered_regions.append(region)

    return filtered_regions


@validate_call
def _insert_agents_into_nearest_regions(
    regions: List[Region],
    agent_properties: List[AgentProperties],
    agent_states: List[AgentState],
    return_region_index: Optional[bool] = False,
    random_seed: Optional[int] = None
) -> Union[List[Region],Tuple[List[Region],List[Tuple[int,int]]]]:
    """
    Helper function to place pre-existing agents into a group of regions. If agents exist 
    within the bounds of multiple regions, it is placed within the region to which whose 
    center it is closest. Agents will be placed "into" the region that is closest even if 
    it is not within the bounds of the region. The length of the agent_states list must be
    equal or less than the length of agent_properties. To remain compliant with :func:`initialize`, 
    agents with defined agent states are placed at the beginning of the list. Optionally 
    using the return_region_index parameter will return a list indicating in which region 
    the agent is placed to preserve agent indexing. A random seed parameter is included for 
    repeatability.

    Arguments
    ----------
    regions:
        A list of Regions with bounds and centre defined for which agents are associated. 

    agent_states:
        Please refer to the documentation of :func:`drive` for information on this parameter.

    agent_properties:
        Please refer to the documentation of :func:`drive` for information on this parameter.

    return_region_index:
        Whether to map the region in which agents of the same index have been placed. Returns 
        a list of the same size as the agent_properties parameter.
    """

    num_agent_states = len(agent_states)
    num_regions = len(regions)
    assert num_regions > 0, "Invalid parameter: number of regions must be greater than zero."
    assert len(agent_properties) >= num_agent_states, "Invalid parameters: number of agent properties must be larger than number agent states."

    if return_region_index: 
        region_map = []
    else:
        region_map = None

    if len(agent_states) > 0: 
        region_agent_states_lengths = [len(region.agent_states) for region in regions]

        for i, (prop, state) in enumerate(zip(agent_properties[:num_agent_states],agent_states)):
            region_distances = []
            for region in regions:
                region_distances.append(sqrt((state.center.x-region.center.x)**2 + (state.center.y-region.center.y)**2))

            closest_region_index = region_distances.index(min(region_distances))
            insert_index = region_agent_states_lengths[closest_region_index]
            region_agent_states_lengths[closest_region_index] += 1
            regions[closest_region_index].agent_properties.insert(insert_index,prop)
            regions[closest_region_index].agent_states.insert(insert_index,state)

            if return_region_index: region_map.append(tuple([closest_region_index,insert_index]))

    if random_seed is not None: seed(random_seed)
    for prop in agent_properties[num_agent_states:]:
        random_region_index = randint(0,num_regions-1)
        regions[random_region_index].agent_properties.append(prop)

        if return_region_index: region_map.append(tuple([random_region_index,len(regions[random_region_index].agent_properties)-1]))

    return regions, region_map


def _consolidate_all_responses(
    all_responses: List[InitializeResponse],
    region_map: Optional[List[Tuple[int,int]]] = None,
    return_exact_agents: bool = False,
    get_infractions: bool = False
):
    if len(all_responses) > 0:
        # Get non-region-specific values such as api_model_version and traffic_light_states from an existing response
        response = deepcopy(all_responses[0]) 

        agent_states = []
        agent_properties = []
        recurrent_states = []
        infractions = []

        region_agent_keep_map = {i: [True]*len(res.agent_properties) for i, res in enumerate(all_responses)}

        if region_map is not None:
            for (region_id, agent_id) in region_map:
                try:
                    agent_states.append(all_responses[region_id].agent_states[agent_id])
                    agent_properties.append(all_responses[region_id].agent_properties[agent_id])
                    recurrent_states.append(all_responses[region_id].recurrent_states[agent_id])

                    if get_infractions:
                        infractions.append(all_responses[region_id].infractions[agent_id])

                    region_agent_keep_map[region_id][agent_id] = False
                except IndexError as e: 
                    exception_message = f"Warning: Unable to fetch specified agent ID {agent_id} in region {region_id}."
                    if not return_exact_agents: 
                        iai.logger.debug(exception_message)
                    else:
                        raise InvertedAIError(message=exception_message)
        
        for ind, response in enumerate(all_responses):
            response_agent_states = response.agent_states
            agent_states = agent_states + [state for i, state in enumerate(response_agent_states) if region_agent_keep_map[ind][i]]
            agent_properties = agent_properties + [prop for i, prop in enumerate(response.agent_properties[:len(response_agent_states)]) if region_agent_keep_map[ind][i]]
            recurrent_states = recurrent_states + [recurr for i, recurr in enumerate(response.recurrent_states) if region_agent_keep_map[ind][i]]
            if get_infractions:
                infractions = infractions + [infr for i, infr in enumerate(response.infractions) if region_agent_keep_map[ind][i]]
        
        response.infractions = infractions
        response.agent_states = agent_states
        response.agent_properties = agent_properties
        response.recurrent_states = recurrent_states 
    else:
        raise InvertedAIError(message=f"Unable to initialize any given region. Please check the input parameters.")

    return response


def _get_all_existing_agents_from_regions(
    regions: List[Region],
    exclude_index: Optional[int] = None,
    nearby_region: Optional[Region] = None,
):
    agent_states = []
    agent_properties = []

    for ind, region in enumerate(regions):
        if not ind == exclude_index:
            if nearby_region is not None:
                if sqrt((nearby_region.center.x-region.center.x)**2+(nearby_region.center.y-region.center.y)**2) > (REGION_MAX_SIZE + AGENT_SCOPE_FOV_BUFFER):
                    continue
            region_agent_states = region.agent_states
            agent_states = agent_states + region_agent_states
            agent_properties = agent_properties + [prop for prop in region.agent_properties[:len(region_agent_states)]]

    return agent_states, agent_properties


def _initialize_regions(
    location: str,
    regions: List[Region],
    traffic_light_state_history: Optional[List[TrafficLightStatesDict]] = None,
    get_infractions: bool = False,
    random_seed: Optional[int] = None,
    api_model_version: Optional[str] = None,
    display_progress_bar: bool = True,
    return_exact_agents: bool = False
) -> Tuple[List[Region],List[InitializeResponse]]:
    
    agent_states_sampled = []
    agent_properties_sampled = []
    agent_rs_sampled = []

    def inside_fov(center: Point, agent_scope_fov: float, point: Point) -> bool:
        return ((center.x - (agent_scope_fov / 2) < point.x < center.x + (agent_scope_fov / 2)) and
                (center.y - (agent_scope_fov / 2) < point.y < center.y + (agent_scope_fov / 2)))
    
    if display_progress_bar:
        iterable_regions = tenumerate(
            regions, 
            total=len(regions),
            desc=f"Initializing regions"
        )
    else:
        iterable_regions = enumerate(regions)

    num_attempts = 1 + len(regions) // ATTEMPT_PER_NUM_REGIONS
    all_responses = []
    for i, region in iterable_regions:
        region_center = region.center
        region_size = region.size

        existing_agent_states, existing_agent_properties = _get_all_existing_agents_from_regions(
            regions = regions,
            exclude_index = i,
            nearby_region = region
        )

        # Acquire agents that exist in other regions that must be passed as conditional to avoid collisions
        out_of_region_conditional_agents = list(filter(
            lambda x: inside_fov(center=region_center, agent_scope_fov=region_size+AGENT_SCOPE_FOV_BUFFER, point=x[0].center), 
            zip(existing_agent_states,existing_agent_properties)
        ))

        out_of_region_conditional_agent_states = [x[0] for x in out_of_region_conditional_agents]
        out_of_region_conditional_agent_properties = [x[1] for x in out_of_region_conditional_agents]

        region_conditional_agent_states = [] if region.agent_states is None else region.agent_states
        num_region_conditional_agents = len(region_conditional_agent_states)
        region_conditional_agent_properties = [] if region.agent_properties is None else region.agent_properties[:num_region_conditional_agents]
        region_unsampled_agent_properties = [] if region.agent_properties is None else region.agent_properties[num_region_conditional_agents:]
        all_agent_states = out_of_region_conditional_agent_states + region_conditional_agent_states
        all_agent_properties = out_of_region_conditional_agent_properties + region_conditional_agent_properties + region_unsampled_agent_properties

        num_out_of_region_conditional_agents = len(out_of_region_conditional_agent_states)

        regions[i].clear_agents()
        response = None
        if len(all_agent_properties) > 0:
            for attempt in range(num_attempts):
                try:
                    response = iai.initialize(
                        location=location,
                        states_history=None if len(all_agent_states) == 0 else [all_agent_states],
                        agent_properties=all_agent_properties,
                        get_infractions=get_infractions,
                        traffic_light_state_history=traffic_light_state_history,
                        location_of_interest=(region_center.x, region_center.y),
                        random_seed=random_seed
                    )

                except InvertedAIError as e:
                    # If error has occurred, display the warning and retry
                    iai.logger.debug(f"Region initialize attempt {attempt} error: {e}")
                    continue

                # Initialization of this region was successful, break the loop and proceed to the next region
                break
            
            else:
                exception_string = f"Unable to initialize region {i} at {region.center} with size {region.size} after {num_attempts} attempts."
                if return_exact_agents: 
                    raise InvertedAIError(message=exception_string)
                else:
                    iai.logger.debug(exception_string)
                    if num_region_conditional_agents > 0:
                    # Get the recurrent states for all predefined agents within the region
                        response = iai.initialize(
                            location=location,
                            states_history=[all_agent_states],
                            agent_properties=all_agent_properties[:num_out_of_region_conditional_agents+num_region_conditional_agents],
                            get_infractions=get_infractions,
                            traffic_light_state_history=traffic_light_state_history,
                            location_of_interest=(region_center.x, region_center.y),
                            random_seed=random_seed
                        )
            
            if response is not None:
                # Filter out conditional agents from other regions
                infractions = []
                for j, (state, props, r_state) in enumerate(zip(
                    response.agent_states[num_out_of_region_conditional_agents:],
                    response.agent_properties[num_out_of_region_conditional_agents:],
                    response.recurrent_states[num_out_of_region_conditional_agents:]
                )):
                    if not return_exact_agents:
                        if not inside_fov(center=region_center, agent_scope_fov=region_size, point=state.center):
                            continue

                    regions[i].insert_all_agent_details(state,props,r_state)
                    if get_infractions:
                        infractions.append(response.infractions[num_out_of_region_conditional_agents:][j])

                response.infractions = infractions
                response.agent_states = regions[i].agent_states
                response.agent_properties = regions[i].agent_properties
                response.recurrent_states = regions[i].recurrent_states
                all_responses.append(response)

                if traffic_light_state_history is None and response.traffic_lights_states is not None:
                    traffic_light_state_history = [response.traffic_lights_states]
        else:
            #There are no agents to initialize within this region, proceed to the next region
            continue

    return regions, all_responses


@validate_call
def large_initialize(
    location: str,
    regions: List[Region],
    agent_properties: Optional[List[AgentProperties]] = None,
    agent_states: Optional[List[AgentState]] = None,
    traffic_light_state_history: Optional[List[TrafficLightStatesDict]] = None,
    get_infractions: bool = False,
    random_seed: Optional[int] = None,
    api_model_version: Optional[str] = None,
    display_progress_bar: bool = True,
    return_exact_agents: bool = False
) -> InitializeResponse:
    """
    A utility function to initialize an area larger than 100x100m. This function takes in a 
    list of Region objects on each of which :func:`initialize` is run and each initialize 
    response is combined into a single response which is returned. While looping over all
    regions, if there are agents in other regions that are near enough to the region of
    interest, they will be passed as conditional to :func:`initialize`. :func:`initialize` 
    will not be called if no agents are specified in the region. 
    A boolean flag can be used to control failure behaviour if :func:`initialize` is unable 
    to produce viable vehicle placements, specifically whether the initialization should 
    continue initializing other regions or raise an exception.

    As well, predefined agents may be passed to this function in 2 different ways. If the index
    of the predefined agents must be preserved, pass these agents' data into the agent_properties 
    and agent_states parameters. Each agent for which its states is defined MUST have its respective
    agent properties defined as well but an agent is permitted to be defined by its properties only
    and :func:`initialize` will fill in the state information. If the index of the predefined agents
    does not matter, they may be placed directly into the region objects or passed into the parameters
    mentioned previously, but make sure to avoid adding these agents more than once.

    Arguments
    ----------
    location:
        Please refer to the documentation of :func:`initialize` for information on this parameter.

    regions:
        List of regions that contains information about the center, size, agent states and 
        agent properties formatted to align with :func:`initialize`. Please refer to the 
        documentation for :func:`initialize` for more details. 

    agent_properties:
        The properties of the agents that will have their indexes preserved within the response 
        output. Please refer to the documentation of :func:`initialize` for information on this 
        parameter.

    agent_states:
        One timestep worth of agent states that will have their indexes preserved within the response 
        output. Please refer to the documentation of :func:`initialize` for more information on this 
        parameter.
    
    traffic_light_state_history:
        Please refer to the documentation of :func:`initialize` for information on this parameter.

    get_infractions:
        Please refer to the documentation of :func:`initialize` for information on this parameter.
    
    random_seed:
        Please refer to the documentation of :func:`initialize` for information on this parameter.

    api_model_version:
        Please refer to the documentation of :func:`initialize` for information on this parameter.

    display_progress_bar:
        If True, a bar is displayed showing the progress of all relevant processes.

    return_exact_agents:
        If set to True, this function will raise an InvertedAIError exception if it cannot fit 
        the requested number of agents in any single region. If set to False, a region that 
        fails to return the number of requested agents will be skipped and only its predefined 
        agents (if any) will be returned with respective RecurrentState's. 
    
    See Also
    --------
    :func:`initialize`
    """

    if (agent_properties is not None and agent_states is not None) or (agent_properties is None and agent_states is not None):
        assert len(agent_properties) >= len(agent_states), "Invalid parameters: number of agent properties must be larger than number agent states."

    regions, region_map = _insert_agents_into_nearest_regions(
        regions = regions,
        agent_properties = [] if agent_properties is None else agent_properties,
        agent_states = [] if agent_states is None else agent_states,
        return_region_index = True,
        random_seed = random_seed
    )

    regions, all_responses = _initialize_regions(
        location = location,
        regions = regions,
        traffic_light_state_history = traffic_light_state_history,
        get_infractions = get_infractions,
        random_seed = random_seed,
        api_model_version = api_model_version,
        display_progress_bar = display_progress_bar,
        return_exact_agents = return_exact_agents
    )

    response = _consolidate_all_responses(
        all_responses = all_responses,
        region_map = region_map,
        return_exact_agents = return_exact_agents,
        get_infractions = get_infractions
    )
    
    return response


import json
import os
import re
import csv
import math
import logging
import random
import time
import numpy as np
import warnings

from typing import Dict, Optional, List, Tuple, Union, Any
from copy import deepcopy
from pydantic import validate_call, validate_arguments

import requests
from requests import Response
from requests.auth import AuthBase
from requests.adapters import HTTPAdapter, Retry

import matplotlib.pyplot as plt
from matplotlib import animation
from matplotlib.patches import Rectangle
from matplotlib.animation import FuncAnimation
from matplotlib.axes import Axes
from matplotlib import transforms

import invertedai as iai
import invertedai.api
import invertedai.api.config
from invertedai import error
from invertedai.future import to_thread
from invertedai.error import InvertedAIError
from invertedai.common import (
    AgentState, 
    AgentAttributes, 
    AgentProperties, 
    AgentType,
    RecurrentState,
    StaticMapActor,
    TrafficLightState, 
    TrafficLightStatesDict 
)

H_SCALE = 10
text_x_offset = 0
text_y_offset = 0.7
text_size = 7
TIMEOUT_SECS = 600
MAX_RETRIES = 10
AGENT_SCOPE_FOV = 120

logger = logging.getLogger(__name__)

STATUS_MESSAGE = {
    403: "Access denied. Please check the provided API key.",
    429: "Throttled",
    502: "The server is having trouble communicating. This is usually a temporary issue. Please try again later.",
    504: "The server took too long to respond. Please try again later.",
    500: "The server encountered an unexpected issue. We're working to resolve this. Please try again later.",
}


class Session:
    def __init__(self,debug_logger=None):
        self.session = requests.Session()
        self.session.mount(
            "https://",
            requests.adapters.HTTPAdapter(),
        )
        self.session.mount(
            "http://",
            requests.adapters.HTTPAdapter(),
        )
        self.session.headers.update(
            {
                "Content-Type": "application/json",
                "Accept-Encoding": "gzip, deflate, br",
                "Connection": "keep-alive",
                "x-client-version": iai.__version__,
            }
        )
        self._base_url = self._get_base_url()
        self._max_retries = float("inf")
        self._status_force_list = [403, 408, 429, 500, 502, 503, 504]
        self._base_backoff = 1  # Base backoff time in seconds
        self._backoff_factor = 2
        self._jitter_factor = 0.5
        self._current_backoff = self._base_backoff
        self._max_backoff = None

        self._debug_logger = debug_logger

    @property
    def base_url(self):
        return self._base_url

    @property
    def max_retries(self):
        return self._max_retries

    @max_retries.setter
    def max_retries(self, value):
        self._max_retries = value

    @property
    def status_force_list(self):
        return self._status_force_list

    @status_force_list.setter
    def status_force_list(self, value):
        self._status_force_list = value.copy()

    @property
    def base_backoff(self):
        return self._base_backoff

    @base_backoff.setter
    def base_backoff(self, value):
        self._base_backoff = value
        self.current_backoff = (
            self._base_backoff
        )  # Reset current_backoff when base_backoff changes

    @property
    def backoff_factor(self):
        return self._backoff_factor

    @backoff_factor.setter
    def backoff_factor(self, value):
        self._backoff_factor = value

    
    @property
    def current_backoff(self):
        return self._current_backoff
    
    @current_backoff.setter
    def current_backoff(self, value):
        self._current_backoff = value

    @property
    def max_backoff(self):
        return self._max_backoff
    
    @max_backoff.setter
    def max_backoff(self, value):
        self._max_backoff = value

    @property
    def jitter_factor(self):
        return self._jitter_factor
    
    @jitter_factor.setter
    def jitter_factor(self, value):
        self._jitter_factor = value


    def should_log(self, retry_count):
        return retry_count == 0 or math.log2(retry_count).is_integer()

    @base_url.setter
    def base_url(self, value):
        self._base_url = value

    def _verify_api_key(
        self, 
        api_token: str, 
        verifying_url: str
    ):
        """
        Verifies the API key by making a request to the verifying URL.

        Args:
            api_token (str): The API token to be used for authentication.
            verifying_url (str): The URL to be used for verification.

        Returns:
            str: The final verifying URL after fallback (if applicable).

        Raises:
            error.AuthenticationError: If access is denied due to an invalid API key.
        """
        self.session.auth = APITokenAuth(api_token)
        response = self.session.request(method="get", url=verifying_url)
        if verifying_url == iai.commercial_url and response.status_code != 200:
            # Check for academic access in case the previous call to the commercial server fails.
            logger.warning(
                "Commercial access denied and fallback to check for academic access."
            )
            verifying_url = iai.academic_url
            response_acd = self.session.request(method="get", url=verifying_url)
            if response_acd.status_code == 200:
                self.base_url = verifying_url
                response = response_acd
            elif response_acd.status_code != 403:
                response = response_acd
        if response.status_code == 403:
            raise error.AuthenticationError(
                "Access denied. Please check the provided API key."
            )
        return verifying_url

    def add_apikey(
        self,
        api_token: str = "",
        key_type: Optional[str] = None,
        url: Optional[str] = None,
    ):
        """
        Bind an API key to the session for authentication.

        Args:
            api_token (str): The API key to be added. Defaults to an empty string.
            key_type (str, optional): The type of API key. Defaults to None. When passed, the base_url will be set according to the key_type.
            url (str, optional): The URL to be used for the request. Defaults to None. When passed, the base_rul will be set to the passed value 
            and the key_type will be ignored.

        Raises:
            InvalidAPIKeyError: If the API key is empty and not in development mode.
            InvalidAPIKeyError: If the key_type is invalid.
            AuthenticationError: If access is denied due to an invalid API key.
            APIError: If the server encounters an error or is unable to perform the requested method.
        """
        if not iai.dev and not api_token:
            raise error.InvalidAPIKeyError("Empty API key received.")
        if url is None:
            request_url = self._get_base_url()
        if key_type is not None and key_type not in ["commercial", "academic"]:
            raise error.InvalidAPIKeyError(f"Invalid API key type: {key_type}.")
        if key_type == "academic":
            request_url = iai.academic_url
        elif key_type == "commercial":
            request_url = iai.commercial_url
        if url is not None:
            request_url = url
        self.base_url = self._verify_api_key(api_token, request_url)

    def use_mock_api(
        self, 
        use_mock: bool = True
    ) -> None:
        invertedai.api.config.mock_api = use_mock
        if use_mock:
            iai.logger.warning(
                "Using mock Inverted AI API - predictions will be trivial"
            )

    async def async_request(
        self, 
        *args, 
        **kwargs
    ):
        return await to_thread(self.request, *args, **kwargs)

    def request(
        self, 
        model: str, 
        params: Optional[dict] = None, 
        data: Optional[dict] = None
    ):
        method, relative_path = iai.model_resources[model]
        
        if self._debug_logger is not None:
            self._debug_logger.append_request(model,data)

        response = self._request(
            method=method,
            relative_path=relative_path,
            params=params,
            json_body=data,
        )

        if self._debug_logger is not None:
            self._debug_logger.append_response(model,response)

        return response

    def _request(
        self,
        method,
        relative_path: str = "",
        params=None,
        headers=None,
        json_body=None,
        data=None,
    ) -> Dict:
        try:
            retries = 0
            while retries < self.max_retries:
                try:
                    response = self.session.request(
                        method=method,
                        params=params,
                        url=self.base_url + relative_path,
                        headers=headers,
                        data=data,
                        json=json_body,
                    )
                except (requests.exceptions.Timeout, requests.exceptions.ConnectionError) as e:
                    logger.warning("Error communicating with IAI, will retry.")
                    response = None
                if response is not None and response.status_code not in self.status_force_list:
                    self.current_backoff = max(
                        self.base_backoff, self.current_backoff / self.backoff_factor
                    )
                    response.raise_for_status()
                    break
                else:
                    if self.jitter_factor is not None:
                        jitter = random.uniform(-self.jitter_factor, self.jitter_factor)
                    else:
                        jitter = 0
                    if self.should_log(retries):
                        if response is not None:
                            logger.warning(
                                f"Retrying {relative_path}: Status {response.status_code}, Message {STATUS_MESSAGE.get(response.status_code, response.text)} Retry #{retries + 1}, Backoff {self.current_backoff} seconds"
                            )
                        else:
                            logger.warning(f"Retrying {relative_path}: No response received, Retry #{retries + 1}, Backoff {self.current_backoff} seconds")
                    time.sleep(min(self.current_backoff * (1 + jitter), self.max_backoff if self.max_backoff is not None else float("inf")))
                    self.current_backoff *= self.backoff_factor
                    if self.max_backoff is not None:
                        self.current_backoff = min(
                            self.current_backoff, self.max_backoff
                        )
                    retries += 1
            else:
                if response is not None:
                    response.raise_for_status()
                else:
                    error.APIConnectionError(
                        "Error communicating with IAI", should_retry=True)


        except requests.exceptions.ConnectionError as e:
            raise error.APIConnectionError(
                "Error communicating with IAI", should_retry=True
            ) from None
        except requests.exceptions.Timeout as e:
            raise error.APIConnectionError("Error communicating with IAI") from None
        except requests.exceptions.RequestException as e:
            if e.response.status_code == 403:
                raise error.AuthenticationError(STATUS_MESSAGE[403]) from None
            elif e.response.status_code in [400, 422]:
                raise error.InvalidRequestError(e.response.text, param="") from None
            elif e.response.status_code == 404:
                raise error.ResourceNotFoundError(e.response.text) from None
            elif e.response.status_code == 408:
                raise error.RequestTimeoutError(e.response.text) from None
            elif e.response.status_code == 413:
                raise error.RequestTooLarge(e.response.text) from None
            elif e.response.status_code == 429:
                raise error.RateLimitError(STATUS_MESSAGE[429]) from None
            elif e.response.status_code == 502:
                raise error.APIError(STATUS_MESSAGE[502]) from None
            elif e.response.status_code == 503:
                raise error.RequestTimeoutError(e.response.text) from None
            elif e.response.status_code == 504:
                raise error.ServiceUnavailableError(STATUS_MESSAGE[504]) from None
            elif 400 <= e.response.status_code < 500:
                raise error.APIError(e.response.text) from None
            else:
                raise error.APIError(STATUS_MESSAGE[500]) from None
        iai.logger.info(
            iai.logger.logfmt(
                "IAI API response",
                path=self.base_url,
                response_code=response.status_code,
            )
        )
        try:
            data = json.loads(response.content)
        except json.decoder.JSONDecodeError:
            raise error.APIError(
                f"HTTP code {response.status_code} from API ({response.content})",
                response.content,
                response.status_code,
                headers=response.headers,
            )
        return data

    def _get_base_url(self) -> str:
        """
        This function returns the endpoint for API calls, which includes the
        version and other endpoint specifications.
        The method path should be appended to the base_url
        """
        if not iai.dev:
            base_url = iai.commercial_url  # Default to commercial when initializing.
        else:
            base_url = iai.dev_url
        # TODO: Add endpoint option and versioning to base_url
        return base_url

    def _handle_error_response(
        self, 
        rbody, 
        rcode, 
        resp, 
        rheaders, 
        stream_error=False
    ):
        try:
            error_data = resp["error"]
        except (KeyError, TypeError):
            raise error.APIError(
                "Invalid response object from API: %r (HTTP response code "
                "was %d)" % (rbody, rcode),
                rbody,
                rcode,
                resp,
            )

        if "internal_message" in error_data:
            error_data["message"] += "\n\n" + error_data["internal_message"]

        iai.logger.info(
            iai.logger.logfmt(
                "IAI API error received",
                error_code=error_data.get("code"),
                error_type=error_data.get("type"),
                error_message=error_data.get("message"),
                error_param=error_data.get("param"),
                stream_error=stream_error,
            )
        )

        # Rate limits were previously coded as 400's with code 'rate_limit'
        if rcode == 429:
            return error.RateLimitError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )
        elif rcode in [400, 404, 415]:
            return error.InvalidRequestError(
                error_data.get("message"),
                error_data.get("param"),
                error_data.get("code"),
                rbody,
                rcode,
                resp,
                rheaders,
            )
        elif rcode == 401:
            return error.AuthenticationError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )
        elif rcode == 403:
            return error.PermissionError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )
        elif rcode == 409:
            return error.TryAgain(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )
        else:
            return error.APIError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )

    def _interpret_response_line(
        self, 
        result
    ):
        rbody = result.content
        rcode = result.status_code
        rheaders = result.headers

        if rcode == 503:
            raise error.ServiceUnavailableError(
                "The server is overloaded or not ready yet.",
                rbody,
                rcode,
                headers=rheaders,
            )
        try:
            data = json.loads(rbody)
        except BaseException:
            raise error.APIError(
                f"HTTP code {rcode} from API ({rbody})", rbody, rcode, headers=rheaders
            )
        if "error" in data or not 200 <= rcode < 300:
            raise self._handle_error_response(rbody, rcode, data, rheaders)

        return data


@validate_call
def get_default_agent_properties(
    agent_count_dict: Dict[AgentType,int],
    use_agent_properties: Optional[bool] = True
) -> List[Union[AgentAttributes,AgentProperties]]:
    """
    Function that outputs a list a AgentAttributes with minimal default settings. 
    Mainly meant to be used to pad a list of AgentAttributes to send as input to
    initialize(). This list is created by reading a dictionary containing the
    desired agent types with the agent count for each type respectively.
    If desired to use deprecate AgentAttributes instead of AgentProperties, set the
    use_agent_properties flag to False.
    """

    agent_attributes_list = []

    for agent_type, agent_count in agent_count_dict.items():
        for _ in range(agent_count):
            if use_agent_properties:
                agent_properties = AgentProperties(agent_type=agent_type)
                agent_attributes_list.append(agent_properties)
            else:
                agent_attributes_list.append(AgentAttributes.fromlist([agent_type]))

    return agent_attributes_list


@validate_call
def convert_attributes_to_properties(
    attributes: AgentAttributes
) -> AgentProperties:
    """
    Convert deprecated AgentAttributes data type to AgentProperties.
    """

    properties = AgentProperties(
        length=attributes.length,
        width=attributes.width,
        rear_axis_offset=attributes.rear_axis_offset,
        agent_type=attributes.agent_type,
        waypoint=attributes.waypoint
    )

    return properties


@validate_call
def iai_conditional_initialize(
    location: str, 
    agent_type_count: Dict[str,int],
    location_of_interest: Tuple[float] = (0,0),
    recurrent_states: Optional[List[RecurrentState]] = None,
    agent_properties: Optional[List[AgentProperties]] = None,
    states_history: Optional[List[List[AgentState]]] = None,
    traffic_light_state_history: Optional[List[TrafficLightStatesDict]] = None, 
    get_birdview: Optional[bool] = False,
    get_infractions: Optional[bool] = False,
    random_seed: Optional[int] = None,
    api_model_version: Optional[str] = None
):
    """
    A utility function to run initialize with conditional agents located at arbitrary distances from the location
    of interest. Only agents within a defined distance of the location of interest are passed to initialize as 
    conditional. Agents outisde of this distance are padded on to the initialize response, including their reccurent
    states. Recurrent states must be provided for all agents, otherwise this function behaves like :func:`initialize`.
    Please refer to the documentation for :func:`initialize` for more information.

    Arguments
    ----------
    location:
        Location name in IAI format.

    agent_type_count:
        A dictionary containing valid AgentType strings as keys mapped to an integer value specifying the desired
        number of agents of that type to initialize.

    location_of_interest:
        Optional coordinates for spawning agents with the given location as center instead of the default map center
    
    See Also
    --------
    :func:`initialize`
    """

    conditional_agent_properties = []
    conditional_agent_states_indexes = []
    conditional_recurrent_states = []
    outside_agent_states = []
    outside_agent_properties = []
    outside_recurrent_states = []

    current_agent_states = states_history[-1]
    conditional_agent_type_count = deepcopy(agent_type_count)
    for i in range(len(current_agent_states)):
        agent_state = current_agent_states[i]
        dist = math.dist(location_of_interest, (agent_state.center.x, agent_state.center.y))
        if dist < AGENT_SCOPE_FOV:
            conditional_agent_states_indexes.append(i)
            conditional_agent_properties.append(agent_properties[i])
            conditional_recurrent_states.append(recurrent_states[i])

            conditional_agent_type = agent_properties[i].agent_type
            if conditional_agent_type in conditional_agent_type_count:
                conditional_agent_type_count[conditional_agent_type] -= 1
                if conditional_agent_type_count[conditional_agent_type] <= 0:
                    del conditional_agent_type_count[conditional_agent_type]

        else:
            outside_agent_states.append(agent_state)
            outside_agent_properties.append(agent_properties[i])
            outside_recurrent_states.append(recurrent_states[i])

    if not conditional_agent_type_count: #The dictionary is empty.
        iai.logger.warning("Agent count requirement already satisfied, no new agents initialized.")

    padded_agent_properties = get_default_agent_properties(conditional_agent_type_count)
    conditional_agent_properties.extend(padded_agent_properties)

    conditional_agent_states = [[]*len(conditional_agent_states_indexes)]
    for ts in range(len(conditional_agent_states)):
        for agent_index in conditional_agent_states_indexes:
            conditional_agent_states[ts].append(states_history[ts][agent_index])

    response = invertedai.api.initialize(
        location = location,
        agent_properties = conditional_agent_properties,
        states_history = conditional_agent_states,
        location_of_interest = location_of_interest,
        traffic_light_state_history = traffic_light_state_history,
        get_birdview = get_birdview,
        get_infractions = get_infractions,
        random_seed = random_seed,
        api_model_version = api_model_version
    )
    response.agent_properties = response.agent_properties + outside_agent_properties
    response.agent_states = response.agent_states + outside_agent_states
    response.recurrent_states = response.recurrent_states + outside_recurrent_states

    return response


class APITokenAuth(AuthBase):
    def __init__(
        self, 
        api_token
    ):
        self.api_token = api_token

    def __call__(
        self, 
        r
    ):
        r.headers["x-api-key"] = self.api_token
        r.headers["api-key"] = self.api_token
        return r


def Jupyter_Render():
    import ipywidgets as widgets
    import matplotlib.pyplot as plt
    import numpy as np

    class Jupyter_Render(widgets.HBox):
        def __init__(self):
            super().__init__()
            output = widgets.Output()
            self.buffer = [np.zeros([128, 128, 3], dtype=np.uint8)]

            with output:
                self.fig, self.ax = plt.subplots(
                    constrained_layout=True, figsize=(5, 5)
                )
            self.im = self.ax.imshow(self.buffer[0])
            self.ax.set_axis_off()

            self.fig.canvas.toolbar_position = "bottom"

            self.max = 0
            # define widgets
            self.play = widgets.Play(
                value=0,
                min=0,
                max=self.max,
                step=1,
                description="Press play",
                disabled=False,
            )
            self.int_slider = widgets.IntSlider(
                value=0, min=0, max=self.max, step=1, description="Frame"
            )

            controls = widgets.HBox(
                [
                    self.play,
                    self.int_slider,
                ]
            )
            controls.layout = self._make_box_layout()
            widgets.jslink((self.play, "value"), (self.int_slider, "value"))
            output.layout = self._make_box_layout()

            self.int_slider.observe(self.update, "value")
            self.children = [controls, output]

        def update(
            self, 
            change
        ):
            self.im.set_data(self.buffer[self.int_slider.value])
            self.fig.canvas.draw()

        def add_frame(
            self, 
            frame
        ):
            self.buffer.append(frame)
            self.int_slider.max += 1
            self.play.max += 1
            self.int_slider.value = self.int_slider.max
            self.play.value = self.play.max

        def _make_box_layout(self):
            return widgets.Layout(
                border="solid 1px black",
                margin="0px 10px 10px 0px",
                padding="5px 5px 5px 5px",
            )

    return Jupyter_Render()


class IAILogger(logging.Logger):
    def __init__(
        self,
        name: str = "IAILogger",
        level: str = "WARNING",
        consoel: bool = True,
        log_file: bool = False,
    ) -> None:

        level = logging.getLevelName(level)
        log_level = level if isinstance(level, int) else 30
        super().__init__(name, log_level)
        if consoel:
            consoel_handler = logging.StreamHandler()
            self.addHandler(consoel_handler)
        if log_file:
            file_handler = logging.FileHandler("iai.log")
            self.addHandler(file_handler)

    @staticmethod
    def logfmt(
        message, 
        **params
    ):
        props = dict(message=message, **params)

        def fmt(key, val):
            # Handle case where val is a bytes or bytesarray
            if hasattr(val, "decode"):
                val = val.decode("utf-8")
            # Check if val is already a string to avoid re-encoding into ascii.
            if not isinstance(val, str):
                val = str(val)
            if re.search(r"\s", val):
                val = repr(val)
            # key should already be a string
            if re.search(r"\s", key):
                key = repr(key)
            return f"{key}={val}"

        return " ".join([fmt(key, val) for key, val in sorted(props.items())])


def rot(rot):
    """Rotate in 2d"""
    return np.array([[np.cos(rot), -np.sin(rot)], [np.sin(rot), np.cos(rot)]])


class ScenePlotter():
    """
    A class providing features for handling the data visualization of a scene involving IAI data.
    """
    def __init__(
        self,
        map_image: Optional[np.array] = None,
        fov: Optional[float] = None,
        xy_offset: Optional[Tuple[float,float]] = None,
        static_actors: Optional[List[StaticMapActor]] = None,
        open_drive: Optional[str] = None, 
        resolution: Tuple[int,int] = (640, 480), 
        dpi: float = 100,
        left_hand_coordinates: bool = False,
        **kwargs
    ):
        """
        Arguments
        ----------
        map_image:
            An image used as the background for the visualization decoded from the birdview map taken from location info.
        fov:
            A single float value representing the field of view of the visualization that can be taken from location info.
        xy_offset:
            A tuple coordinate of the center of the map in metres that can be taken from location info.
        static_actors:
            A list of StaticMapActor objects representing objects such as traffic lights that can be taken from location info.
        open_drive: 
            If using an ASAM OpenDRIVE format map for visualization, this string parameter is used to indicate the path to the corresponding CSV file.
        resolution: 
            The desired resolution of the map image expressed as a Tuple with two integers for the width and height respectively.
        dpi:
            Dots per inch to define the level of detail in the image.
        left_hand_coordinates:
            Boolean flag dictating whether the X-coordinates of all agents and actors should be reversed to fit a left hand coordinate system.

        Keyword Arguments
        -----------------
        map_image:
            Base image onto which the scene is visualized. This parameter must be provided if using an ASAM OpenDRIVE format map.
        fov: float
            The field of view in meters corresponding to the map_image attribute. This parameter must be provided if using an ASAM OpenDRIVE format map.
        xy_offset:
            The left-hand offset for the center of the map image. This parameter must be provided if using an ASAM OpenDRIVE format map.
        static_actors:
            A list of static actor agents (e.g. traffic lights) represented as StaticMapActor objects, in the scene. This parameter must be provided
             if using an ASAM OpenDRIVE format map.
        
        See Also
        --------
        :func:`location_info`
        """

        self._left_hand_coordinates = left_hand_coordinates

        self.conditional_agents = None
        self.agent_properties = None
        self.traffic_lights_history = None
        self.agent_states_history = None
        
        self._open_drive = open_drive
        self._dpi = dpi
        self._resolution = resolution
        
        self.map_image = map_image
        self.fov = fov
        self.xy_offset = xy_offset
        self.static_actors = static_actors

        self.traffic_lights = {static_actor.actor_id: static_actor for static_actor in self.static_actors if static_actor.agent_type == 'traffic_light'}

        if self._open_drive is None:
            self.extent = (- self.fov / 2 + self.xy_offset[0], self.fov / 2 + self.xy_offset[0]) + \
                (- self.fov / 2 + self.xy_offset[1], self.fov / 2 + self.xy_offset[1])

        self.traffic_light_colors = {
            "red": (1.0, 0.0, 0.0),
            "green": (0.0, 1.0, 0.0),
            "yellow": (1.0, 0.8, 0.0),
        }

        self.agent_c = (0.125,0.29,0.529)
        self.agent_ped_c = (1.0, 0.75, 0.8)
        self.cond_c = (0.78, 0.0, 0.0)
        self.dir_c = (0.392,1.0,1.0)
        self.v_c = (0.2, 0.75, 0.2)

        self.dir_lines = {}
        self.v_lines = {}
        self.actor_boxes = {}
        self.traffic_light_boxes = {}
        self.box_labels = {}
        self.frame_label = None
        self.current_ax = None

        self.numbers = None

        self.agent_face_colors = None 
        self.agent_edge_colors = None 

        self.reset_recording()

    @validate_arguments
    def initialize_recording(
        self,
        agent_states: List[AgentState], 
        agent_attributes: Optional[List[AgentAttributes]] = None, 
        agent_properties: Optional[List[AgentProperties]] = None,
        traffic_light_states: Optional[Dict[int, TrafficLightState]] = None, 
        conditional_agents: Optional[List[int]] = None
    ):
        """
        Record the initial state of the scene to be visualized. This function also acts as an implicit reset of the recording and removes previous 
        agent state, agent attribute, conditional agent, traffic light, and agent style data.

        Arguments
        ----------
        agent_states:
            A list of AgentState objects corresponding to the initial time step to be visualized.
        agent_attributes:
            Static attributes of the agent, which don’t change over the course of a simulation. We assume every agent is a rectangle obeying a 
            kinematic bicycle model.
        agent_properties:
            Static attributes of the agent (with the AgentProperties data type), which don’t change over the course of a simulation. We assume every 
            agent is a rectangle obeying a kinematic bicycle model.
        traffic_light_states:
            Optional parameter containing the state of the traffic lights corresponding to the initial time step to be visualized. This parameter 
            should only be used if the corresponding map contains traffic light static actors.
        conditional_agents:
            Optional parameter containing a list of agent IDs corresponding to conditional agents to be visualized to distinguish themselves.
        """

        assert (agent_attributes is not None) ^ (agent_properties is not None), \
            "Either agent_attributes or agent_properties is populated. Populating both or neither field is invalid."

        if agent_attributes is not None:
            self.agent_properties = [convert_attributes_to_properties(attr) for attr in agent_attributes]
            warnings.warn('agent_attributes is deprecated. Please use agent_properties.',category=DeprecationWarning)
        else:
            self.agent_properties = agent_properties

        self.agent_states_history = [agent_states]
        self.traffic_lights_history = [traffic_light_states]
        if conditional_agents is not None:
            self.conditional_agents = conditional_agents
        else:
            self.conditional_agents = []

        self.agent_face_colors = None
        self.agent_edge_colors = None

    def reset_recording(self):
        """
        Explicitly reset the recording and remove the previous agent state, agent attribute, conditional agent, traffic light, and agent style data.
        """
        self.agent_states_history = []
        self.traffic_lights_history = []
        self.agent_properties = None
        self.conditional_agents = []
        self.agent_properties = None
        self.agent_face_colors = None 
        self.agent_edge_colors = None 

    @validate_arguments
    def record_step(
        self,
        agent_states: List[AgentState], 
        traffic_light_states: Optional[Dict[int, TrafficLightState]] = None
    ):
        """
        Record a single timestep of scene data to be used in a visualization

        Arguments
        ----------
        agent_states:
            A list of AgentState objects corresponding to the initial time step to be visualized.
        traffic_light_states:
            Optional parameter containing the state of the traffic lights corresponding to the initial time step to be visualized. This parameter should
            only be used if the corresponding map contains traffic light static actors.
        """
        self.agent_states_history.append(agent_states)
        self.traffic_lights_history.append(traffic_light_states)

    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def plot_scene(
        self,
        agent_states: List[AgentState], 
        agent_attributes: Optional[List[AgentAttributes]] = None, 
        agent_properties: Optional[List[AgentProperties]] = None, 
        traffic_light_states: Optional[Dict[int, TrafficLightState]] = None, 
        conditional_agents: Optional[List[int]] = None,
        ax: Optional[Axes] = None,
        numbers: Optional[List[int]] = None, 
        direction_vec: bool = True, 
        velocity_vec: bool = False,
        agent_face_colors: Optional[List[Optional[Tuple[float,float,float]]]] = None,
        agent_edge_colors: Optional[List[Optional[Tuple[float,float,float]]]] = None
    ):
        """
        Plot a single timestep of data then reset the recording. 

        Arguments
        ----------
        agent_states:
            A list of agents to be visualized in the image.
        agent_attributes: 
            Static attributes of the agent, which don’t change over the course of a simulation. We assume every agent is a rectangle obeying a kinematic
            bicycle model.
        agent_properties:
            Static attributes of the agent (with the AgentProperties data type), which don’t change over the course of a simulation. We assume every 
            agent is a rectangle obeying a kinematic bicycle model.
        traffic_light_states: 
            Optional parameter containing the state of the traffic lights to be visualized in the image. This parameter should only be used if the 
            corresponding map contains traffic light static actors.
        conditional_agents:
            Optional parameter containing a list of agent IDs of conditional agents to be visualized in the image to distinguish themselves.
        ax: 
            A matplotlib Axes object used to plot the image. By default, an Axes object is created if a value of None is passed.
        numbers: 
            A list of agent ID's that should be plotted in the image. By default this value is set to None.
        direction_vec:
            Flag to determine if a vector showing the vehicles direction should be plotted in the image. By default this flag is set to True.
        velocity_vec: 
            Flag to determine if the a vector showing the vehicles velocity should be plotted in the animation. By default this flag is set to False.
        agent_face_colors:
            An optional parameter containing a list of either RGB tuples indicating the desired color of the agent with the corresponding index ID. A value 
            of None in this list will use the default color. This value gets overwritten by the conditional agent color.
        agent_edge_colors:
            An optional parameter containing a list of either RGB tuples indicating the desired color of a border around the agent with the corresponding 
            index ID. A value of None in this list will use the default color. This value gets overwritten by the conditional agent color.

        """
        assert (agent_attributes is not None) ^ (agent_properties is not None), \
            "Either agent_attributes or agent_properties is populated. Populating both or neither field is invalid."

        if agent_attributes is not None:
            agent_properties = [convert_attributes_to_properties(attr) for attr in agent_attributes]
            warnings.warn('agent_attributes is deprecated. Please use agent_properties.',category=DeprecationWarning)

        self.initialize_recording(
            agent_states=agent_states, 
            agent_properties=agent_properties,
            traffic_light_states=traffic_light_states,
            conditional_agents=conditional_agents
        )

        self._validate_agent_style_data(
            agent_face_colors=agent_face_colors,
            agent_edge_colors=agent_edge_colors
        )

        self._plot_frame(
            idx=0, 
            ax=ax, 
            numbers=numbers, 
            direction_vec=direction_vec,
            velocity_vec=velocity_vec, 
            plot_frame_number=False
        )

        self.reset_recording()

    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def animate_scene(
        self,
        output_name: Optional[str] = None,
        start_idx: int = 0, 
        end_idx: int = -1,
        ax: Optional[Axes] = None,
        numbers: Optional[List[int]] = None, 
        direction_vec: bool = True, 
        velocity_vec: bool = False,
        plot_frame_number: bool = False, 
        agent_face_colors: Optional[List[Optional[Tuple[float,float,float]]]] = None,
        agent_edge_colors: Optional[List[Optional[Tuple[float,float,float]]]] = None
    ) -> FuncAnimation:
        """
        Produce an animation of sequentially recorded steps. A matplotlib animation object can be returned and/or a gif saved of the scene.

        Parameters
        ----------
        output_name: 
            File name of the gif to which the animation will be saved.
        start_idx:
            The index of the time step from which the animation will begin. By default it is assumed all recorded steps are desired to be animated.
        end_idx:
            The index of the time step from which the animation will end. By default it is assumed all recorded steps are desired to be animated.
        ax: 
            A matplotlib Axes object used to plot the animation. By default, an Axes object is created if a value of None is passed.
        numbers: 
            A list of agent ID's that should be plotted in the image. By default this value is set to None.
        direction_vec: 
            Flag to determine if a vector showing the vehicles direction should be plotted in the animation. By default this flag is set to True.
        velocity_vec:
            Flag to determine if the a vector showing the vehicles velocity should be plotted in the animation. By default this flag is set to False.
        plot_frame_number: 
            Flag to determine if the frame numbers should be plotted in the animation. By default this flag is set to False.
        agent_face_colors:
            An optional parameter containing a list of either RGB tuples indicating the desired color of the agent with the corresponding index ID. A value 
            of None in this list will use the default color. This value gets overwritten by the conditional agent color.
        agent_edge_colors:
            An optional parameter containing a list of either RGB tuples indicating the desired color of a border around the agent with the corresponding index 
            ID. A value of None in this list will use the default color. This value gets overwritten by the conditional agent color.
        """

        self._validate_agent_style_data(agent_face_colors,agent_edge_colors)

        self._initialize_plot(ax=ax, numbers=numbers, direction_vec=direction_vec,
                              velocity_vec=velocity_vec, plot_frame_number=plot_frame_number)
        end_idx = len(self.agent_states_history) if end_idx == -1 else end_idx
        fig = self.current_ax.figure
        fig.set_size_inches(self._resolution[0] / self._dpi, self._resolution[1] / self._dpi, True)

        def animate(i):
            self._update_frame_to(i)

        ani = FuncAnimation(
            fig, animate, np.arange(start_idx, end_idx), interval=100)
        if output_name is not None:
            ani.save(f'{output_name}', writer='pillow', dpi=self._dpi)
        return ani

    def _transform_point_to_left_hand_coordinate_frame(self,x,orientation):
        t_x = 2*self.xy_offset[0] - x
        if orientation >= 0:
            t_orientation = -orientation + math.pi
        else:
            t_orientation = -orientation - math.pi

        return t_x, t_orientation

    def _plot_frame(self, idx, ax=None, numbers=None, direction_vec=True,
                   velocity_vec=False, plot_frame_number=False):
        self._initialize_plot(ax=ax, numbers=numbers, direction_vec=direction_vec,
                              velocity_vec=velocity_vec, plot_frame_number=plot_frame_number)
        self._update_frame_to(idx)

    def _validate_agent_style_data(self,agent_face_colors,agent_edge_colors):
        if self.agent_properties is not None: 
            if agent_face_colors is not None:
                if len(agent_face_colors) != len(self.agent_properties):
                    raise Exception("Number of agent face colors does not match number of agents.")
            if agent_edge_colors is not None:
                if len(agent_edge_colors) != len(self.agent_properties):
                    raise Exception("Number of agent edge colors does not match number of agents.")

        self.agent_face_colors = agent_face_colors
        self.agent_edge_colors = agent_edge_colors

    def _initialize_plot(self, ax=None, numbers=None, direction_vec=True,
                         velocity_vec=False, plot_frame_number=False):
        if ax is None:
            plt.clf()
            ax = plt.gca()
        if self._open_drive is None:
            ax.imshow(self.map_image, extent=self.extent)
        else:
            self._draw_xodr_map(ax)
            self.extent = (self.xy_offset[0] - self.fov / 2, self.xy_offset[0] + self.fov / 2) +\
                (self.xy_offset[1] - self.fov / 2, self.xy_offset[1] + self.fov / 2)

            ax.set_xlim((self.extent[0], self.extent[1]))
            ax.set_ylim((self.extent[2], self.extent[3]))
        self.current_ax = ax

        self.dir_lines = {}
        self.v_lines = {}
        self.actor_boxes = {}
        self.traffic_light_boxes = {}
        self.box_labels = {}
        self.frame_label = None

        self.numbers = numbers
        self.direction_vec = direction_vec
        self.velocity_vec = velocity_vec
        self.plot_frame_number = plot_frame_number

        self._update_frame_to(0)

    def _get_color(self,agent_idx,color_list):
        c = None
        if color_list and color_list[agent_idx]:
            is_good_color_format = isinstance(color_list[agent_idx],tuple)
            for pc in color_list[agent_idx]:
                is_good_color_format *= isinstance(pc,float) and (0.0 <= pc <= 1.0)
            
            if not is_good_color_format:
                raise Exception(f"Expected color format is Tuple[float,float,float] with 0 <= float <= 1 but received {color_list[agent_idx]}.")
            c = color_list[agent_idx]

        return c

    def _update_frame_to(self, frame_idx):
        for i, (agent, agent_attribute) in enumerate(
            zip(self.agent_states_history[frame_idx], self.agent_properties)
        ):
            self._update_agent(i, agent, agent_attribute)

        if self.traffic_lights_history[frame_idx] is not None:
            for light_id, light_state in self.traffic_lights_history[frame_idx].items():
                self._plot_traffic_light(light_id, light_state)

        if self.plot_frame_number:
            if self.frame_label is None:
                self.frame_label = self.current_ax.text(
                    self.extent[0], 
                    self.extent[2], 
                    str(frame_idx), 
                    c="r", 
                    fontsize=18
                )
            else:
                self.frame_label.set_text(str(frame_idx))

        if self._open_drive is None:
            self.current_ax.set_xlim(*self.extent[0:2])
            self.current_ax.set_ylim(*self.extent[2:4])

    def _update_agent(self, agent_idx, agent, agent_attribute):
        l, w = agent_attribute.length, agent_attribute.width
        if agent_attribute.agent_type == "pedestrian":
            l, w = 1.5, 1.5
        x, y = agent.center.x, agent.center.y
        v = agent.speed
        psi = agent.orientation

        if self._left_hand_coordinates:
            x, psi = self._transform_point_to_left_hand_coordinate_frame(x,psi)

        box = np.array([
            [0, 0], [l * 0.5, 0],  # direction vector
            [0, 0], [v * 0.5, 0],  # speed vector at (0.5 m / s ) / m
        ])

        box = np.matmul(rot(psi), box.T).T + np.array([[x, y]])
        if self.direction_vec:
            marker_offset = agent_attribute.length/4
            x_data = x + marker_offset*math.cos(psi)
            y_data = y + marker_offset*math.sin(psi)
            marker_data = (3, 0, (-90+180*psi/math.pi))

            if agent_idx not in self.dir_lines:
                self.dir_lines[agent_idx] = self.current_ax.plot(
                    x_data,
                    y_data,
                    marker=marker_data,
                    markersize=agent_attribute.width*400/self.fov, 
                    linestyle='None',
                    c=self.dir_c
                )
            else:
                self.dir_lines[agent_idx][0].set_xdata(x_data)
                self.dir_lines[agent_idx][0].set_ydata(y_data)
                self.dir_lines[agent_idx][0].set_marker(marker_data)

        if self.velocity_vec:
            if agent_idx not in self.v_lines:
                self.v_lines[agent_idx] = self.current_ax.plot(
                    box[2:4, 0], 
                    box[2:4, 1], 
                    lw=1.5, 
                    c=self.v_c
                )[0]  # plot the speed
            else:
                self.v_lines[agent_idx].set_xdata(box[2:4, 0])
                self.v_lines[agent_idx].set_ydata(box[2:4, 1])
        
        if self.numbers is not None and agent_idx in self.numbers:
            if agent_idx not in self.box_labels:
                self.box_labels[agent_idx] = self.current_ax.text(
                    x, 
                    y, 
                    str(agent_idx), 
                    c="r", 
                    fontsize=18
                )
                self.box_labels[agent_idx].set_clip_on(True)
            else:
                self.box_labels[agent_idx].set_x(x)
                self.box_labels[agent_idx].set_y(y)

        lw = 1
        fc = self._get_color(agent_idx,self.agent_face_colors)
        if fc is None:
            if agent_idx in self.conditional_agents:
                fc = self.cond_c
            else:
                fc = self.agent_c
        ec = self._get_color(agent_idx,self.agent_edge_colors)
        if ec is None:
            lw = 0
            ec = fc

        rect = Rectangle(
            (x - l / 2, y - w / 2), 
            l, 
            w, 
            angle=psi * 180 / np.pi, 
            rotation_point='center', 
            fc=fc, 
            ec=ec, 
            lw=lw
        )

        if agent_idx in self.actor_boxes:
            self.actor_boxes[agent_idx].remove()
        self.actor_boxes[agent_idx] = rect
        self.actor_boxes[agent_idx].set_clip_on(True)
        self.current_ax.add_patch(self.actor_boxes[agent_idx])

    def _plot_traffic_light(self, light_id, light_state):
        light = self.traffic_lights[light_id]
        x, y = light.center.x, light.center.y
        psi = light.orientation
        l, w = max(light.length,1.0), max(light.width,1.0)

        if self._left_hand_coordinates:
            x, psi = self._transform_point_to_left_hand_coordinate_frame(x,psi)

        rect = Rectangle(
            (x - l / 2, y - w / 2),
            l,
            w,
            angle=psi * 180 / np.pi,
            rotation_point="center",
            fc=self.traffic_light_colors[light_state],
            lw=0,
        )
        if light_id in self.traffic_light_boxes:
            self.traffic_light_boxes[light_id].remove()
        self.current_ax.add_patch(rect)
        self.traffic_light_boxes[light_id] = rect

    def _draw_xodr_map(self, ax, extras=False):
        """
        This function plots the parsed xodr map
        the `odrplot` of `esmini` is used for plotting and parsing xodr
        https: // esmini.github.io/  # _tools_overview
        """
        with open(self._open_drive) as f:
            reader = csv.reader(f, skipinitialspace=True)
            positions = list(reader)

        ref_x = []
        ref_y = []
        ref_z = []
        ref_h = []

        lane_x = []
        lane_y = []
        lane_z = []
        lane_h = []

        border_x = []
        border_y = []
        border_z = []
        border_h = []

        road_id = []
        road_id_x = []
        road_id_y = []

        road_start_dots_x = []
        road_start_dots_y = []

        road_end_dots_x = []
        road_end_dots_y = []

        lane_section_dots_x = []
        lane_section_dots_y = []

        arrow_dx = []
        arrow_dy = []

        current_road_id = None
        current_lane_id = None
        current_lane_section = None
        new_lane_section = False

        for i in range(len(positions) + 1):

            if i < len(positions):
                pos = positions[i]

            # plot road id before going to next road
            if i == len(positions) or (
                pos[0] == "lane" and i > 0 and current_lane_id == "0"
            ):

                if current_lane_section == "0":
                    road_id.append(int(current_road_id))
                    index = int(len(ref_x[-1]) / 3.0)
                    h = ref_h[-1][index]
                    road_id_x.append(
                        ref_x[-1][index]
                        + (text_x_offset * math.cos(h) - text_y_offset * math.sin(h))
                    )
                    road_id_y.append(
                        ref_y[-1][index]
                        + (text_x_offset * math.sin(h) + text_y_offset * math.cos(h))
                    )
                    road_start_dots_x.append(ref_x[-1][0])
                    road_start_dots_y.append(ref_y[-1][0])
                    if len(ref_x) > 0:
                        arrow_dx.append(ref_x[-1][1] - ref_x[-1][0])
                        arrow_dy.append(ref_y[-1][1] - ref_y[-1][0])
                    else:
                        arrow_dx.append(0)
                        arrow_dy.append(0)

                lane_section_dots_x.append(ref_x[-1][-1])
                lane_section_dots_y.append(ref_y[-1][-1])

            if i == len(positions):
                break

            if pos[0] == "lane":
                current_road_id = pos[1]
                current_lane_section = pos[2]
                current_lane_id = pos[3]
                if pos[3] == "0":
                    ltype = "ref"
                    ref_x.append([])
                    ref_y.append([])
                    ref_z.append([])
                    ref_h.append([])

                elif pos[4] == "no-driving":
                    ltype = "border"
                    border_x.append([])
                    border_y.append([])
                    border_z.append([])
                    border_h.append([])
                else:
                    ltype = "lane"
                    lane_x.append([])
                    lane_y.append([])
                    lane_z.append([])
                    lane_h.append([])
            else:
                if ltype == "ref":
                    ref_x[-1].append(float(pos[0]))
                    ref_y[-1].append(float(pos[1]))
                    ref_z[-1].append(float(pos[2]))
                    ref_h[-1].append(float(pos[3]))

                elif ltype == "border":
                    border_x[-1].append(float(pos[0]))
                    border_y[-1].append(float(pos[1]))
                    border_z[-1].append(float(pos[2]))
                    border_h[-1].append(float(pos[3]))
                else:
                    lane_x[-1].append(float(pos[0]))
                    lane_y[-1].append(float(pos[1]))
                    lane_z[-1].append(float(pos[2]))
                    lane_h[-1].append(float(pos[3]))

        # plot driving lanes in blue
        for i in range(len(lane_x)):
            ax.plot(lane_x[i], lane_y[i], linewidth=1.0, color="#222222")

        # plot road ref line segments
        for i in range(len(ref_x)):
            ax.plot(ref_x[i], ref_y[i], linewidth=2.0, color="#BB5555")

        # plot border lanes in gray
        for i in range(len(border_x)):
            ax.plot(border_x[i], border_y[i], linewidth=1.0, color="#AAAAAA")

        if extras:
            # plot red dots indicating lane dections
            for i in range(len(lane_section_dots_x)):
                ax.plot(
                    lane_section_dots_x[i],
                    lane_section_dots_y[i],
                    "o",
                    ms=4.0,
                    color="#BB5555",
                )

            for i in range(len(road_start_dots_x)):
                # plot a yellow dot at start of each road
                ax.plot(
                    road_start_dots_x[i],
                    road_start_dots_y[i],
                    "o",
                    ms=5.0,
                    color="#BBBB33",
                )
                # and an arrow indicating road direction
                ax.arrow(
                    road_start_dots_x[i],
                    road_start_dots_y[i],
                    arrow_dx[i],
                    arrow_dy[i],
                    width=0.1,
                    head_width=1.0,
                    color="#BB5555",
                )
            # plot road id numbers
            for i in range(len(road_id)):
                ax.text(
                    road_id_x[i],
                    road_id_y[i],
                    road_id[i],
                    size=text_size,
                    ha="center",
                    va="center",
                    color="#3333BB",
                )

        return None




from pydantic import BaseModel, validate_arguments
from typing import List, Optional, Dict, Tuple

import matplotlib.pyplot as plt
import json

from invertedai import location_info
from invertedai.utils import ScenePlotter
from invertedai.api.location import LocationResponse
from invertedai.api.initialize import InitializeResponse
from invertedai.api.drive import DriveResponse
from invertedai.common import ( 
    AgentAttributes, 
    AgentProperties,
    AgentState, 
    LightRecurrentState,
    LightRecurrentStates,
    Point,
    RecurrentState,
    TrafficLightStatesDict 
)


class ScenarioLog(BaseModel):
    """
    A log containing simulation information for storage, replay, or an initial state from which a simulation 
    can be continued. Some data fields contain data for all historic time steps while others contain information
    for the most recent time step to be used to continue a simulation.
    """

    agent_states: List[List[AgentState]] #: Historic data for all agents states up until the most recent time step.
    agent_properties: List[AgentProperties] #: Agent properties data for all agents in this scenario/log.
    traffic_lights_states: Optional[List[TrafficLightStatesDict]] = None #: Historic data for all TrafficLightStatesDict up until the most recent time step.

    location: str #: Location name in IAI format.
    rendering_center: Optional[Tuple[float, float]] = None #: Please refer to the documentation of :func:`location_info` for information on this parameter.
    rendering_fov: Optional[int] = None #: Please refer to the documentation of :func:`location_info` for information on this parameter.

    lights_random_seed: Optional[int] = None #: Controls the stochastic aspects of the the traffic lights states.
    initialize_random_seed: Optional[int] = None #: Please refer to the documentation of :func:`initialize` for information on the random_seed parameter.
    drive_random_seed: Optional[int] = None #: Please refer to the documentation of :func:`drive` for information on the random_seed parameter.

    initialize_model_version: Optional[str] = "best" #: Please refer to the documentation of :func:`initialize` for information on the api_model_version parameter.
    drive_model_version: Optional[str] = "best" #: Please refer to the documentation of :func:`drive` for information on the api_model_version parameter.
    
    light_recurrent_states: Optional[LightRecurrentStates] = None #: As of the most recent time step. Please refer to the documentation of :func:`drive` for further information on this parameter.
    recurrent_states: Optional[List[RecurrentState]] = None #: As of the most recent time step. Please refer to the documentation of :func:`drive` for further information on this parameter.

    waypoints: Optional[Dict[str,List[Point]]] = None #: As of the most recent time step. A list of waypoints keyed to agent ID's not including waypoints already passed. These waypoints are not automatically populated into the agent properties.


class LogBase():
    """
    A class for containing features relevant to both log reading and writing such as visualization.
    """

    def __init__(self):
        self._scenario_log = None
        self.simulation_length = None

    @validate_arguments
    def visualize_range(
        self,
        timestep_range: Tuple[int,int],
        gif_path: str,
        fov: int = 200,
        resolution: Tuple[int,int] = (2048,2048),
        dpi: int = 300,
        map_center: Optional[Tuple[float,float]] = None,
        direction_vec: bool = False,
        velocity_vec: bool = False,
        plot_frame_number: bool = True,
        left_hand_coordinates: bool = False
    ):
        """
        Use the available internal tools to visualize the a specific range of time steps within the log and save it to a given location. If
        an invalid time step range is given, the function will fail. Please refer to ScenePlotter for details on the visualization tool.
        """

        for timestep in timestep_range:
            assert timestep >= 0 or timestep <= (self.simulation_length - 1), "Visualization time range valid."
        assert timestep_range[1] >= timestep_range[0], "Visualization time range valid."

        location_info_response = location_info(
            location=self._scenario_log.location,
            rendering_fov=fov,
            rendering_center=map_center
        )
        rendered_static_map = location_info_response.birdview_image.decode()
        map_center = tuple([location_info_response.map_center.x, location_info_response.map_center.y]) if map_center is None else map_center

        scene_plotter = ScenePlotter(
            map_image=rendered_static_map,
            fov=fov,
            xy_offset=map_center,
            static_actors=location_info_response.static_actors,
            resolution=resolution,
            dpi=dpi,
            left_hand_coordinates=left_hand_coordinates
        )
        scene_plotter.initialize_recording(
            agent_states=self._scenario_log.agent_states[timestep_range[0]],
            agent_properties=self._scenario_log.agent_properties,
            traffic_light_states=None if self._scenario_log.traffic_lights_states is None else self._scenario_log.traffic_lights_states[timestep_range[0]]
        )

        traffic_lights_states = [None]*(timestep_range[1]-timestep_range[0]-1) if self._scenario_log.traffic_lights_states is None else self._scenario_log.traffic_lights_states[timestep_range[0]:timestep_range[1]]
        for states, lights in zip(self._scenario_log.agent_states[1:timestep],traffic_lights_states):
            scene_plotter.record_step(states,lights)

        fig, ax = plt.subplots(constrained_layout=True, figsize=(50, 50))
        plt.axis('off')
        scene_plotter.animate_scene(
            output_name=gif_path,
            ax=ax,
            direction_vec=direction_vec,
            velocity_vec=velocity_vec,
            plot_frame_number=plot_frame_number
        )

        plt.close(fig)

    @validate_arguments
    def visualize(
        self,
        gif_path: str,
        fov: int = 200,
        resolution: Tuple[int,int] = (2048,2048),
        dpi: int = 300,
        map_center: Optional[Tuple[float,float]] = None,
        direction_vec: bool = False,
        velocity_vec: bool = False,
        plot_frame_number: bool = True,
        left_hand_coordinates: bool = False
    ):
        """
        Use the available internal tools to visualize the entire log and save it to a given location. Please refer to ScenePlotter for details on 
        the visualization tool.
        """

        self.visualize_range(
            timestep_range = tuple([0,self.simulation_length-1]),
            gif_path = gif_path,
            fov = fov,
            resolution = resolution,
            dpi = dpi,
            map_center = map_center,
            direction_vec = direction_vec,
            velocity_vec = velocity_vec,
            plot_frame_number = plot_frame_number
        )

    def initialize(self):
        pass

    def drive(self):
        pass


class LogWriter(LogBase):
    """
    A class for conveniently writing a log to a JSON log format. 
    """

    def __init__(self):
        super().__init__()

    @validate_arguments
    def export_to_file(
        self,
        log_path: str,
        scenario_log: Optional[ScenarioLog] = None
    ):  
        """
        Convert the data currently contained within the log into a JSON format and export it to a given file. This function can furthermore be 
        used to export a given scenario log instead of the log contained within the object.
        """

        if scenario_log is None:
            scenario_log = self._scenario_log
        num_cars, num_pedestrians = 0, 0
        for prop in scenario_log.agent_properties:
            if prop.agent_type == "car":
                num_cars += 1
            elif prop.agent_type == "pedestrian":
                num_pedestrians += 1

        num_controls_light, num_controls_yield, num_controls_stop, num_controls_other = 0, 0, 0, 0
        static_actors_list = location_info(location=scenario_log.location).static_actors
        for actor in static_actors_list:
            if actor.agent_type == "traffic_light":
                num_controls_light += 1
            elif actor.agent_type == "yield_sign":
                num_controls_yield += 1
            elif actor.agent_type == "stop_sign":
                num_controls_stop += 1
            else:
                num_controls_other += 1

        predetermined_agents_dict = {}
        for i, prop in enumerate(scenario_log.agent_properties):
            states_dict = {}
            for t, states in enumerate(scenario_log.agent_states):
                states_dict[str(t)] = {
                    "center": {"x": states[i].center.x, "y": states[i].center.y},
                    "orientation": states[i].orientation,
                    "speed": states[i].speed
                }

            predetermined_agents_dict[str(i)] = {
                "entity_type": prop.agent_type,
                "static_attributes": {
                    "length": prop.length,
                    "width": prop.width,
                    "rear_axis_offset": prop.rear_axis_offset,
                },
                "states":states_dict
            }

        predetermined_controls_dict = {}
        if scenario_log.traffic_lights_states is not None:
            for actor in [actor for actor in static_actors_list if actor.agent_type == "traffic_light"]:
                actor_id = actor.actor_id
                states_dict = {}

                for t, tls in enumerate(scenario_log.traffic_lights_states):
                    states_dict[str(t)] = {
                        "center": {"x": actor.center.x, "y": actor.center.y},
                        "orientation": actor.orientation,
                        "speed": 0,
                        "control_state": tls[actor_id]
                    }

                predetermined_controls_dict[actor_id] = {
                    "entity_type": "traffic_light",
                    "static_attributes": {
                        "length": actor.length,
                        "width": actor.width,
                        "rear_axis_offset": 0,
                    },
                    "states":states_dict
                }

        individual_suggestions_dict = {}
        for i, prop in enumerate(scenario_log.agent_properties):
            wp = prop.waypoint
            if wp is not None:
                individual_suggestions_dict[str(i)] = {
                    "suggestion_strength": 0.8,
                    "states": {
                        "0": {
                            "center": {
                                "x": wp.x,
                                "y": wp.y
                            }
                        }
                    }
                }

        self.output_dict = {
            "location": {
                "identifier": scenario_log.location
            },
            "scenario_length": len(scenario_log.agent_states),
            "num_agents": {
                "car": num_cars,
                "pedestrian": num_pedestrians
            },
            "predetermined_agents": predetermined_agents_dict,
            "num_controls": {
                "traffic_light": num_controls_light,
                "yield_sign": num_controls_yield,
                "stop_sign": num_controls_stop,
                "other": num_controls_other,
            },
            "predetermined_controls": predetermined_controls_dict,
            "individual_suggestions": individual_suggestions_dict,
            "initialize_random_seed": scenario_log.initialize_random_seed,
            "lights_random_seed": scenario_log.lights_random_seed,
            "drive_random_seed": scenario_log.drive_random_seed,
            "drive_model_version": scenario_log.drive_model_version,
            "initialize_model_version": scenario_log.initialize_model_version,
            "birdview_options": {
                "rendering_center": [
                    scenario_log.rendering_center[0],
                    scenario_log.rendering_center[1]
                ],
                "renderingFOV": scenario_log.rendering_fov
            },
            "light_recurrent_states": [] if scenario_log.light_recurrent_states is None else [lrs.tolist() for lrs in scenario_log.light_recurrent_states],
            "rendering_centers": [
                scenario_log.rendering_center[0],
                scenario_log.rendering_center[1]
            ]
        }

        with open(log_path, "w") as outfile:
            json.dump(self.output_dict, outfile)

    @classmethod
    def export_log_to_file(
        cls, 
        log_path: str,
        scenario_log: ScenarioLog
    ):
        """
        Class function to convert a given log data type into a JSON format and export it to a given file.
        """

        cls.export_to_file(cls,log_path,scenario_log)

    @validate_arguments
    def initialize(
        self,
        location: str,
        location_info_response: LocationResponse,
        init_response: InitializeResponse,
        lights_random_seed: Optional[int] = None,
        initialize_random_seed: Optional[int] = None,
        drive_random_seed: Optional[int] = None
    ): 
        """
        Consume and store all initial information within a ScenarioLog data object. If random seed information is desired to be stored, it 
        must be given separately but is not mandatory.
        """

        agent_properties = init_response.agent_properties
        if type(agent_properties[0]) == AgentAttributes:
            agent_properties = [convert_attributes_to_properties(attr) for attr in agent_properties]

        self._scenario_log = ScenarioLog(
            agent_states=[init_response.agent_states], 
            agent_properties=agent_properties, 
            traffic_lights_states=[init_response.traffic_lights_states] if init_response.traffic_lights_states is not None else None, 
            location=location,
            rendering_center=[
                location_info_response.map_center.x,
                location_info_response.map_center.y
            ],
            rendering_fov=location_info_response.map_fov,
            lights_random_seed=lights_random_seed,
            initialize_random_seed=initialize_random_seed,
            drive_random_seed=drive_random_seed,
            initialize_model_version=init_response.api_model_version,
            drive_model_version=None,
            light_recurrent_states=init_response.light_recurrent_states,
            recurrent_states=init_response.recurrent_states,
            waypoints=None
        )

        self.simulation_length = 1

    @validate_arguments
    def drive(
        self,
        drive_response: DriveResponse
    ): 
        """
        Consume and store driving response information from a single timestep and append it to the end of the log.  
        """

        self._scenario_log.agent_states.append(drive_response.agent_states)
        if drive_response.traffic_lights_states is not None:
            self._scenario_log.traffic_lights_states.append(drive_response.traffic_lights_states)
        
        self._scenario_log.drive_model_version = drive_response.api_model_version
        self._scenario_log.light_recurrent_states = drive_response.light_recurrent_states
        self._scenario_log.recurrent_states = drive_response.recurrent_states

        self.simulation_length += 1


class LogReader(LogBase):
    """
    A class for conveniently reading in a log file then rendering it and/or plugging it into a simulation. Once the log is read, it is 
    intended to be used in place of calling the API.
    """

    def __init__(
        self,
        log_path: str
    ):
        """
        The initialization of this object must be given the path to a JSON file in the IAI format. Assume that the 0th time step is taken
        from the output of :func:`initialize` and set the time step to the 1st time step whic correlates to the first time step produced 
        by :func:`drive`.
        """

        super().__init__()

        with open(log_path) as f:
            LOG_DATA = json.load(f)

        location = LOG_DATA["location"]["identifier"]

        all_agent_states = []
        all_agent_properties = []
        for i in range(LOG_DATA["scenario_length"]):
            agent_states_ts = []
            for agent_num, agent in LOG_DATA["predetermined_agents"].items():
                if i == 0:
                    agent_attributes_json = agent["static_attributes"]
                    agent_properties = AgentProperties()
                    agent_properties.length = agent_attributes_json["length"]
                    agent_properties.width = agent_attributes_json["width"]
                    agent_properties.rear_axis_offset = agent_attributes_json["rear_axis_offset"]
                    agent_properties.agent_type = agent["entity_type"]
                    all_agent_properties.append(agent_properties)

                agent_state = agent["states"][str(i)]
                agent_states_ts.append(AgentState.fromlist([
                    agent_state["center"]["x"],
                    agent_state["center"]["y"],
                    agent_state["orientation"],
                    agent_state["speed"],
                ]))

            all_agent_states.append(agent_states_ts)

        all_traffic_light_states = []
        for i in range(LOG_DATA["scenario_length"]):
            traffic_light_states_ts = {}
            for actor_id, actor in LOG_DATA["predetermined_controls"].items():
                if actor["entity_type"] == "traffic_light":
                    actor_info_ts = actor["states"][str(i)]
                    traffic_light_states_ts[int(actor_id)] = actor_info_ts["control_state"]
            if traffic_light_states_ts:
                all_traffic_light_states.append(traffic_light_states_ts)

        if not all_traffic_light_states:
            all_traffic_light_states = None

        agent_waypoints = {}
        for agent_id, waypoints in LOG_DATA["individual_suggestions"].items():
            agent_waypoints[agent_id] = []
            for i, pt in waypoints["states"].items():
                data = pt["center"]
                agent_waypoints[agent_id].append(Point.fromlist([data["x"],data["y"]]))
        if not agent_waypoints:
            agent_waypoints = None

        self._scenario_log = ScenarioLog(
            agent_states=all_agent_states, 
            agent_properties=all_agent_properties, 
            traffic_lights_states=all_traffic_light_states, 
            location=location, 
            rendering_center=tuple([LOG_DATA["birdview_options"]["rendering_center"][0],LOG_DATA["birdview_options"]["rendering_center"][1]]),
            rendering_fov=LOG_DATA["birdview_options"]["renderingFOV"],
            lights_random_seed=LOG_DATA["lights_random_seed"],
            initialize_random_seed=LOG_DATA["initialize_random_seed"],
            drive_random_seed=LOG_DATA["drive_random_seed"],
            initialize_model_version=LOG_DATA["initialize_model_version"],
            drive_model_version=LOG_DATA["drive_model_version"],
            light_recurrent_states=None if LOG_DATA["light_recurrent_states"] is [] else [LightRecurrentState(state=state[0],time_remaining=state[1]) for state in LOG_DATA["light_recurrent_states"]],
            recurrent_states=None,
            waypoints=agent_waypoints
        )
        self._scenario_log_original = self._scenario_log

        self.reset_log()

        self.simulation_length = len(all_agent_states)
        self.location = location
        self.initialize_model_version = self._scenario_log.initialize_model_version
        self.drive_model_version = self._scenario_log.drive_model_version
        
        self.location_info_response = location_info(
            location=self._scenario_log.location,
            rendering_fov=self._scenario_log.rendering_fov,
            rendering_center=self._scenario_log.rendering_center,
        )

    @validate_arguments
    def _return_state_at_timestep(
        self,
        timestep: int
    ):
        """
        Populate all state data from the given time step into the relevant member variables. 
        """

        if timestep >= self.simulation_length:
            return False

        self.agent_states = self._scenario_log.agent_states[timestep]
        self.recurrent_states = None
        self.traffic_lights_states = None if self._scenario_log.traffic_lights_states is None else self._scenario_log.traffic_lights_states[timestep]
        self.light_recurrent_states = self._scenario_log.light_recurrent_states if timestep == (self.simulation_length - 1) else None

        return True

    @validate_arguments
    def return_last_state(self):
        """
        Read and make available state data from the final time step contained within the log which is useful as a launching point for another simulation.
        """

        return self._return_state_at_timestep(timestep=self.simulation_length-1)

    @validate_arguments
    def initialize(self):
        """
        Read and make available state data from the 0th time step into the relevant state member variables e.g. agent_states. Furthermore, set the 
        agent_properties state variable here analogously to how :func:`initialize` returns this information through the API.
        """

        self.agent_properties = self._scenario_log.agent_properties
        is_init_response = self._return_state_at_timestep(timestep=0)
        self.current_timestep = 1

        return is_init_response

    @validate_arguments
    def drive(self):
        """
        Read and make available state data from the current time step into the relevant member variables then increment the current time step so that this 
        function may be called again. If the end of the log has been reached, return False otherwise return True.
        """

        if self.current_timestep >= self.simulation_length:
            return False

        is_drive_response = self._return_state_at_timestep(timestep=self.current_timestep)
        self.current_timestep += 1

        return is_drive_response

    @validate_arguments
    def reset_log(self):
        """
        In the case the log was modified, revert the log to its initial state after being read and clear all state data. Furthermore, change the current 
        time step such that the first :func:`drive` time step can be read.
        """
        
        self._scenario_log = self._scenario_log_original

        self.agent_states = None
        self.agent_properties = None
        self.recurrent_states = None
        self.traffic_lights_states = None
        self.light_recurrent_states = None
        self.current_timestep = 1





import logging
import json
import os

from collections import defaultdict
from typing import List, Optional, Dict, Tuple
from datetime import datetime, timezone

logger = logging.getLogger(__name__)

class DebugLogger:
    def __init__(
        self,
        debug_log_path: str
    ):
        self.debug_log_path = debug_log_path
        self._create_directory()

        self.data = defaultdict(list)

        file_name = "iai_log_" + self._get_current_time_human_readable_UTC() + "_UTC.json"
        self.log_path = self.debug_log_path + file_name

    def _get_current_time_human_readable_UTC(self):
        return datetime.now(timezone.utc).strftime("%Y-%m-%d_%H:%M:%S:%f")

    def _create_directory(self):
        if not os.path.isdir(self.debug_log_path):
            logger.info(f"Debug log directory does not exist: Created new directory at {self.debug_log_path}")
            os.makedirs(self.debug_log_path)

    def append_request(
        self,
        model: str,
        data_dict: Optional[dict] = None
    ):
        ts = self._get_current_time_human_readable_UTC()
        data_str = json.dumps(data_dict)

        if model == "location_info":
            self.data["location_info_requests"].append(data_str)
            self.data["location_info_request_timestamps"].append(ts)

        elif model == "initialize":
            self.data["initialize_requests"].append(data_str)
            self.data["initialize_request_timestamps"].append(ts)

        elif model == "drive":
            self.data["drive_requests"].append(data_str)
            self.data["drive_request_timestamps"].append(ts)

        self.write_data_to_log()

    def append_response(
        self,
        model: str,
        data_dict: Optional[dict] = None
    ):
        ts = self._get_current_time_human_readable_UTC()
        data_str = json.dumps(data_dict)

        if model == "location_info":
            self.data["location_info_responses"].append(data_str)
            self.data["location_info_response_timestamps"].append(ts)

        elif model == "initialize":
            self.data["initialize_responses"].append(data_str)
            self.data["initialize_response_timestamps"].append(ts)

        elif model == "drive":
            self.data["drive_responses"].append(data_str)
            self.data["drive_response_timestamps"].append(ts)

        self.write_data_to_log()

    def write_data_to_log(self):
        with open(self.log_path, "w") as outfile:
            json.dump(self.data, outfile)


from typing import List, Optional, Tuple, Union
from pydantic import BaseModel

from invertedai.common import AgentAttributes, AgentProperties, AgentState, RecurrentState, Point
from invertedai.utils import convert_attributes_to_properties

REGION_MAX_SIZE = 100.0

class Region(BaseModel):
    """
    A region in a map used to divide a large simulation into smaller parts.

    See Also
    --------
    AgentProperties
    """

    center: Point #: The center of the region if such a concept is relevant (e.g. center of a square, center of a rectangle)
    size: float #: Side length of the region for the default interpretation of a region as a square
    agent_states: Optional[List[AgentState]] = [] #: A list of existing agents within the region
    agent_properties: Optional[List[AgentProperties]] = [] #: The static parameters of agents that exist within the region or that will be initialized within the region
    recurrent_states: Optional[List[RecurrentState]] = [] #: Recurrent states of the agents eixsting within the region

    @classmethod
    def create_square_region(
        cls, 
        center: Point, 
        size: Optional[float] = REGION_MAX_SIZE, 
        agent_states: Optional[List[AgentState]] = [], 
        agent_properties: Optional[List[Union[AgentAttributes,AgentProperties]]] = [], 
        recurrent_states: Optional[List[RecurrentState]] = []
    ):
        cls.center = center
        cls.size = size
        for agent in agent_states:
            assert cls.is_inside(cls,agent.center), f"Existing agent states at position {agent.center} must be located within the region."

        
        agent_properties_new = []
        for properties in agent_properties:
            properties_new = properties
            if isinstance(properties,AgentAttributes):
                properties_new = convert_attributes_to_properties(properties_new)
            agent_properties_new.append(properties_new)
        agent_properties = agent_properties_new

        return cls(
            center=center, 
            size=size,
            agent_states=agent_states, 
            agent_properties=agent_properties, 
            recurrent_states=recurrent_states 
        )

    @classmethod
    def copy(
        cls,
        region
    ):

        return cls(
            center=region.center, 
            size=region.size,
            agent_states=region.agent_states, 
            agent_properties=region.agent_properties, 
            recurrent_states=region.recurrent_states 
        )

    def clear_agents(self):

        self.agent_states = []
        self.agent_properties = []
        self.recurrent_states = []

    def insert_all_agent_details(
        self,
        agent_state: AgentState,
        agent_properties: Union[AgentAttributes,AgentProperties],
        recurrent_state: RecurrentState
    ):
        if isinstance(agent_properties,AgentAttributes):
            agent_properties = convert_attributes_to_properties(agent_properties)

        self.agent_states.append(agent_state)
        self.agent_properties.append(agent_properties)
        self.recurrent_states.append(recurrent_state)

    def is_inside(
        self,
        point: Point
    ) -> bool:
        """
        Helper function to check if a point is within an X-Y axis aligned bounding box around the region.
        This function should be faster but equivalent in result to the other checking function if the 
        region is rectangular.

        """

        x, y = point.x, point.y
        region_x, region_y = self.center.x, self.center.y
        return region_x - self.size/2 <= x and x <= region_x + self.size/2 and region_y - self.size/2 <= y and y <= region_y + self.size/2



from typing import Optional

from pydantic import BaseModel

import invertedai as iai
from invertedai.large.common import Region
from invertedai.common import Point, AgentState, AgentProperties, RecurrentState

BUFFER_FOV = 35
QUADTREE_SIZE_BUFFER = 1


class QuadTreeAgentInfo(BaseModel):
    """
    All information relevant to a single agent.

    See Also
    --------
    AgentState
    AgentProperties
    RecurrentState
    """

    agent_state: AgentState
    recurrent_state: Optional[RecurrentState] = None
    agent_properties: AgentProperties
    agent_id: int

    def tolist(self):
        return [self.agent_state, self.agent_properties, self.recurrent_state, self.agent_id]

    @classmethod
    def fromlist(cls, l):
        agent_state, agent_properties, recurrent_state, agent_id = l
        return cls(agent_state=agent_state, agent_properties=agent_properties, recurrent_state=recurrent_state, agent_id=agent_id)


class QuadTree:
    def __init__(
        self, 
        capacity: int, 
        region: Region, 
    ):
        self.capacity = capacity
        self.region = region
        self.leaf = True
        self.northWest = None
        self.northEast = None
        self.southWest = None
        self.southEast = None

        self.region_buffer = Region.create_square_region(
            center=self.region.center,
            size=self.region.size+2*BUFFER_FOV
        )
        self.particles = []
        self.particles_buffer = []

    def subdivide(self):
        parent = self.region
        new_size = self.region.size/2
        new_center_dist = new_size/2
        parent_x = parent.center.x
        parent_y = parent.center.y

        region_nw = Region.create_square_region(
            center=Point.fromlist([parent_x-new_center_dist,parent_y+new_center_dist]),
            size=new_size
        )
        region_ne = Region.create_square_region(
            center=Point.fromlist([parent_x+new_center_dist,parent_y+new_center_dist]),
            size=new_size
        )
        region_sw = Region.create_square_region(
            center=Point.fromlist([parent_x-new_center_dist,parent_y-new_center_dist]),
            size=new_size
        )
        region_se = Region.create_square_region(
            center=Point.fromlist([parent_x+new_center_dist,parent_y-new_center_dist]),
            size=new_size
        )

        self.northWest = QuadTree(self.capacity,region_nw)
        self.northEast = QuadTree(self.capacity,region_ne)
        self.southWest = QuadTree(self.capacity,region_sw)
        self.southEast = QuadTree(self.capacity,region_se)

        self.leaf = False
        self.region.clear_agents()
        self.region_buffer.clear_agents()

        for particle in self.particles:
            is_inserted = self.insert_particle_in_leaf_nodes(particle,False)
        for particle in self.particles_buffer:
            is_inserted = self.insert_particle_in_leaf_nodes(particle,True)
        self.particles = []
        self.particles_buffer = []
        
    def insert_particle_in_leaf_nodes(self,particle,is_inserted):
        is_inserted_in_this_branch = self.northWest.insert(particle,is_inserted)
        is_inserted_in_this_branch = self.northEast.insert(particle,is_inserted_in_this_branch or is_inserted) or is_inserted_in_this_branch
        is_inserted_in_this_branch = self.southWest.insert(particle,is_inserted_in_this_branch or is_inserted) or is_inserted_in_this_branch
        is_inserted_in_this_branch = self.southEast.insert(particle,is_inserted_in_this_branch or is_inserted) or is_inserted_in_this_branch

        return is_inserted_in_this_branch

    def insert(self, particle, is_particle_placed=False):
        is_in_region = self.region.is_inside(particle.agent_state.center)
        is_in_buffer = self.region_buffer.is_inside(particle.agent_state.center)

        if (not is_in_region) and (not is_in_buffer):
            return False

        if (len(self.particles) + len(self.particles_buffer)) < self.capacity and self.leaf:
            if is_in_region and not is_particle_placed:
                self.particles.append(particle)
                self.region.insert_all_agent_details(*particle.tolist()[:-1])
                return True

            else: # Particle is within the buffer region of this leaf node
                self.particles_buffer.append(particle)
                self.region_buffer.insert_all_agent_details(*particle.tolist()[:-1])
                return False

        else:
            if self.leaf:
                self.subdivide()

            is_inserted = self.insert_particle_in_leaf_nodes(particle,is_particle_placed)

            return is_inserted

    def get_regions(self):
        if self.leaf:
            return [self.region]
        else:
            return self.northWest.get_regions() + self.northEast.get_regions() + \
                self.southWest.get_regions() + self.southEast.get_regions()

    def get_leaf_nodes(self):
        if self.leaf:
            return [self]
        else:
            return self.northWest.get_leaf_nodes() + self.northEast.get_leaf_nodes() + \
                self.southWest.get_leaf_nodes() + self.southEast.get_leaf_nodes()

    def get_number_of_agents_in_node(self):
        return len(self.particles)

def _flatten_and_sort(nested_list,index_list):
    flat_list = [x for sublist in nested_list for x in sublist]
    sorted_list = [x[1] for x in sorted(zip(index_list, flat_list))]
    
    return sorted_list


import time
from typing import List, Optional, Tuple, Dict
from pydantic import BaseModel, validate_call

import invertedai as iai
from invertedai.api.config import TIMEOUT, should_use_mock_api
from invertedai.api.mock import (
    get_mock_birdview,
    get_mock_agents_at_fault,
    get_mock_blamed_reasons,
    get_mock_confidence_score,
)
from invertedai.error import APIConnectionError, InvalidInput
from invertedai.common import (
    AgentState,
    Image,
    AgentAttributes,
    TrafficLightStatesDict,
)

class BlameResponse(BaseModel):
    """
    Response returned from an API call to :func:`iai.blame`.
    """
    agents_at_fault: Optional[Tuple[int, ...]] #: A tuple containing all agents predicted to be at fault. If empty, BLAME has predicated no agents are at fault.
    reasons: Optional[Dict[int, List[str]]] #: A dictionary with agent IDs as keys corresponding to "agents_at_fault" paired with a list of reasons why the keyed agent is at fault (e.g. traffic_light_violation).
    confidence_score: Optional[float] #: Float value between [0,1] indicating BLAME's confidence in the response where 0.0 represents the minimum confidence and 1.0 represents maximum.
    birdviews: Optional[List[Image]]  #: If `get_birdviews` was set, this contains the resulting image.

@validate_call
def blame(
    location: str,
    colliding_agents: Tuple[int, int],
    agent_state_history: List[List[AgentState]],
    agent_attributes: List[AgentAttributes],
    traffic_light_state_history: Optional[List[TrafficLightStatesDict]] = None,
    get_reasons: bool = False,
    get_confidence_score: bool = False,
    get_birdviews: bool = False
) -> BlameResponse:
    """
    Parameters
    ----------
    location:
        Location name in IAI format.

    colliding_agents:
        Two agents involved in the collision. These integers should correspond to the 
        indices of the relevant agents in the lists within agent_state_history.

    agent_state_history:
        Lists containing AgentState objects for every agent within the scene (up to 100
        agents) for each time step within the relevant continuous sequence immediately 
        preceding the collision.
        The list of AgentState objects should include the first time step of the collision
        and no time steps afterwards. The lists of AgentState objects preceding the 
        collision should capture enough of the scenario context before the collision for 
        BLAME to analyze and assign fault. For best results it is recommended to 
        input 20-50 time steps of 0.1s each preceding the collision.
        Each AgentState state must include x: [float], y: [float] coordinates in meters, 
        orientation: [float] in radians with 0 pointing along the positive x axis and 
        pi/2 pointing along the positive y axis, and speed: [float] in m/s.

    agent_attributes:
        List of static AgentAttribute objects for all agents. Each agent requires, length: 
        [float], width: [float], and rear_axis_offset: [float] all in meters.

    traffic_light_state_history:
        List of TrafficLightStatesDict objects containing the state of all traffic lights
        for every time step. The dictionary keys are the traffic_light IDs and value is 
        the state, i.e., 'green', 'yellow', 'red', or None.

    get_reasons:
        Whether to return the reasons regarding why each agent was blamed.

    get_confidence_score:
        Whether to return how confident the BLAME is in its response.

    get_birdviews:
        Whether to return the image visualizing the collision case. This is very slow and 
        should only be used for debugging.

    See Also
    --------
    :func:`drive`
    :func:`initialize`
    :func:`location_info`
    :func:`light`
    """
    if len(agent_state_history[0]) != len(agent_attributes):
        raise InvalidInput("Incompatible Number of Agents in either 'agent_state_history' or 'agent_attributes'.")

    if should_use_mock_api():
        agents_at_fault = get_mock_agents_at_fault()
        birdviews = [get_mock_birdview()]
        reasons = get_mock_blamed_reasons()
        confidence_score = get_mock_confidence_score()
        response = BlameResponse(
            agents_at_fault=agents_at_fault,
            birdviews=birdviews,
            reasons=reasons,
            confidence_score=confidence_score
        )
        return response

    model_inputs = dict(
        location=location,
        colliding_agents=colliding_agents,
        agent_state_history=[[state.tolist() for state in agent_states] for agent_states in agent_state_history],
        agent_attributes=[attr.tolist() for attr in agent_attributes],
        traffic_light_state_history=traffic_light_state_history,
        get_reasons=get_reasons ,
        get_confidence_score=get_confidence_score,
        get_birdviews=get_birdviews
    )
    start = time.time()
    timeout = TIMEOUT

    while True:
        try:
            response = iai.session.request(model="blame", data=model_inputs)

            response = BlameResponse(
                agents_at_fault=response["agents_at_fault"],
                reasons=response["reasons"],
                confidence_score=response["confidence_score"],
                birdviews=[Image.fromval(birdview) for birdview in response["birdviews"]]
            )

            return response
        except APIConnectionError as e:
            iai.logger.warning("Retrying")
            if (
                timeout is not None and time.time() > start + timeout
            ) or not e.should_retry:
                raise e


@validate_call
async def async_blame(
    location: str,
    colliding_agents: Tuple[int, int],
    agent_state_history: List[List[AgentState]],
    agent_attributes: List[AgentAttributes],
    traffic_light_state_history: Optional[List[TrafficLightStatesDict]] = None,
    get_reasons: bool = False,
    get_confidence_score: bool = False,
    get_birdviews: bool = False
) -> BlameResponse:
    """
    A light async version of :func:`blame`
    """
    model_inputs = dict(
        location=location,
        colliding_agents=colliding_agents,
        agent_state_history=[[state.tolist() for state in agent_states] for agent_states in agent_state_history],
        agent_attributes=[attr.tolist() for attr in agent_attributes],
        traffic_light_state_history=traffic_light_state_history,
        get_reasons=get_reasons,
        get_confidence_score=get_confidence_score,
        get_birdviews=get_birdviews
    )

    response = await iai.session.async_request(model="blame", data=model_inputs)

    response = BlameResponse(
        agents_at_fault=response["agents_at_fault"],
        reasons=response["reasons"],
        confidence_score=response["confidence_score"],
        birdviews=[Image.fromval(birdview) for birdview in response["birdviews"]]
    )

    return response



import asyncio
import warnings
from typing import Tuple, Optional, List, Union
from pydantic import BaseModel, validate_call
from math import ceil

import invertedai as iai
from invertedai.large.common import Region
from invertedai.common import Point, AgentState, AgentAttributes, AgentProperties, RecurrentState, TrafficLightStatesDict, LightRecurrentState
from invertedai.api.drive import DriveResponse
from invertedai.utils import convert_attributes_to_properties
from invertedai.error import InvertedAIError, InvalidRequestError
from ._quadtree import QuadTreeAgentInfo, QuadTree, _flatten_and_sort, QUADTREE_SIZE_BUFFER

DRIVE_MAXIMUM_NUM_AGENTS = 100

async def async_drive_all(async_input_params):
    all_responses = await asyncio.gather(*[iai.async_drive(**input_params) for input_params in async_input_params])
    return all_responses


@validate_call
def large_drive(
    location: str,
    agent_states: List[AgentState],
    agent_properties: List[Union[AgentAttributes,AgentProperties]],
    recurrent_states: Optional[List[RecurrentState]] = None,
    traffic_lights_states: Optional[TrafficLightStatesDict] = None,
    light_recurrent_states: Optional[List[LightRecurrentState]] = None,
    get_infractions: bool = False,
    random_seed: Optional[int] = None,
    api_model_version: Optional[str] = None,
    single_call_agent_limit: Optional[int] = None,
    async_api_calls: bool = True
) -> DriveResponse:
    """
    A utility function to drive more than the normal capacity of agents in a call to :func:`drive`.
    The agents are inserted into a quadtree structure and :func:`drive` is then called on each
    region represented by a leaf node of the quadtree. Agents near this region are included in the
    :func:`drive` calls to ensure the agents see all their neighbours. The quadtree is constructed 
    during each call to this utility function to maintain statelessness.

    Parameters
    ----------
    location:
        Please refer to the documentation of :func:`drive` for information on this parameter.

    agent_states:
        Please refer to the documentation of :func:`drive` for information on this parameter.

    agent_properties:
        Please refer to the documentation of :func:`drive` for information on this parameter.

    recurrent_states:
        Please refer to the documentation of :func:`drive` for information on this parameter.

    traffic_lights_states:
       Please refer to the documentation of :func:`drive` for information on this parameter.

    light_recurrent_states:
       Please refer to the documentation of :func:`drive` for information on this parameter.

    get_infractions:
        Please refer to the documentation of :func:`drive` for information on this parameter.

    random_seed:
        Please refer to the documentation of :func:`drive` for information on this parameter.

    api_model_version:
        Please refer to the documentation of :func:`drive` for information on this parameter.
    
    single_call_agent_limit:
        The number of agents allowed in a region before it must subdivide. Currently this value 
        represents the capacity of a quadtree leaf node that will subdivide if the number of vehicles 
        in the region, plus relevant neighbouring regions, passes this threshold. In any case, this 
        will be limited to the maximum currently supported by :func:`drive`.

    async_api_calls:
        A flag to control whether to use asynchronous DRIVE calls.

    See Also
    --------
    :func:`drive`
    """

    # Validate input arguments
    if single_call_agent_limit is None:
        single_call_agent_limit = DRIVE_MAXIMUM_NUM_AGENTS
    if single_call_agent_limit > DRIVE_MAXIMUM_NUM_AGENTS:
        single_call_agent_limit = DRIVE_MAXIMUM_NUM_AGENTS
        iai.logger.warning(f"Single Call Agent Limit cannot be more than {DRIVE_MAXIMUM_NUM_AGENTS}, limiting this value to {DRIVE_MAXIMUM_NUM_AGENTS} and proceeding.")
    num_agents = len(agent_states)
    if not (num_agents == len(agent_properties)):
        if recurrent_states is not None and not (num_agents == len(recurrent_states)):
            raise InvalidRequestError(message="Input lists are not of equal size.")
    if not num_agents > 0:
        raise InvalidRequestError(message="Valid call must contain at least 1 agent.")

    # Convert any AgentAttributes to AgentProperties for backwards compatibility 
    agent_properties_new = []
    is_using_attributes = False
    for properties in agent_properties:
        properties_new = properties
        if isinstance(properties,AgentAttributes):
            properties_new = convert_attributes_to_properties(properties)
            is_using_attributes = True
        agent_properties_new.append(properties_new)
    agent_properties = agent_properties_new

    if is_using_attributes:
        warnings.warn('agent_attributes is deprecated. Please use agent_properties.',category=DeprecationWarning)

    # Generate quadtree
    agent_x = [agent.center.x for agent in agent_states]
    agent_y = [agent.center.y for agent in agent_states]
    max_x, min_x, max_y, min_y = max(agent_x), min(agent_x), max(agent_y), min(agent_y)
    region_size = ceil(max(max_x - min_x, max_y - min_y)) + QUADTREE_SIZE_BUFFER
    region_center = (round((max_x+min_x)/2),round((max_y+min_y)/2))

    quadtree = QuadTree(
        capacity=single_call_agent_limit,
        region=Region.create_square_region(
            center=Point.fromlist(list(region_center)),
            size=region_size
        ),
    )
    for i, (agent, attrs) in enumerate(zip(agent_states,agent_properties)):
        if recurrent_states is None:
            recurr_state = None
        else:
            recurr_state = recurrent_states[i]

        agent_info = QuadTreeAgentInfo.fromlist([agent, attrs, recurr_state, i])
        is_inserted = quadtree.insert(agent_info)

        if not is_inserted:
            raise InvertedAIError(message=f"Unable to insert agent into region.")

    
    # Call DRIVE API on all leaf nodes
    all_leaf_nodes = quadtree.get_leaf_nodes()
    async_input_params = []
    all_responses = []
    non_empty_nodes = []
    agent_id_order = []
    
    if len(all_leaf_nodes) > 1:
        for i, leaf_node in enumerate(all_leaf_nodes):
            region, region_buffer = leaf_node.region, leaf_node.region_buffer
            region_agents_ids = [particle.agent_id for particle in leaf_node.particles]

            if len(region.agent_states) > 0:
                non_empty_nodes.append(leaf_node)
                agent_id_order.extend(region_agents_ids)
                input_params = {
                    "location":location,
                    "agent_states":region.agent_states+region_buffer.agent_states,
                    "recurrent_states":None if recurrent_states is None else region.recurrent_states+region_buffer.recurrent_states,
                    "agent_properties":region.agent_properties+region_buffer.agent_properties,
                    "light_recurrent_states":light_recurrent_states,
                    "traffic_lights_states":traffic_lights_states,
                    "get_birdview":False,
                    "rendering_center":None,
                    "rendering_fov":None,
                    "get_infractions":get_infractions,
                    "random_seed":random_seed,
                    "api_model_version":api_model_version
                }
                if not async_api_calls:
                    all_responses.append(iai.drive(**input_params))
                else:
                    async_input_params.append(input_params)

        if async_api_calls:
            all_responses = asyncio.run(async_drive_all(async_input_params))

        response = DriveResponse(
            agent_states = _flatten_and_sort([region_response.agent_states[:leaf_node.get_number_of_agents_in_node()] for region_response, leaf_node in zip(all_responses,non_empty_nodes)],agent_id_order),
            recurrent_states = _flatten_and_sort([region_response.recurrent_states[:leaf_node.get_number_of_agents_in_node()] for region_response, leaf_node in zip(all_responses,non_empty_nodes)],agent_id_order),
            is_inside_supported_area = _flatten_and_sort([region_response.is_inside_supported_area[:leaf_node.get_number_of_agents_in_node()] for region_response, leaf_node in zip(all_responses,non_empty_nodes)],agent_id_order),
            infractions = [] if not get_infractions else _flatten_and_sort([region_response.infractions[:leaf_node.get_number_of_agents_in_node()] for region_response, leaf_node in zip(all_responses,non_empty_nodes)],agent_id_order),
            api_model_version = all_responses[0].api_model_version,
            birdview = None,
            traffic_lights_states = all_responses[0].traffic_lights_states,
            light_recurrent_states = all_responses[0].light_recurrent_states
        )

    else:
        # Quadtree capacity has not been surpassed therefore can just call regular drive()
        response = iai.drive(
            location = location,
            agent_states = agent_states,
            agent_properties = agent_properties,
            recurrent_states = recurrent_states,
            traffic_lights_states = traffic_lights_states,
            light_recurrent_states = light_recurrent_states,
            get_birdview = False,
            rendering_center = None,
            rendering_fov = None,
            get_infractions = get_infractions,
            random_seed = random_seed,
            api_model_version = api_model_version
        )

    return response


import time
import asyncio
import warnings
from pydantic import BaseModel, validate_call
from typing import List, Optional, Dict, Tuple

import invertedai as iai
from invertedai.api.config import TIMEOUT, should_use_mock_api
from invertedai.error import TryAgain, InvalidInputType, InvalidInput
from invertedai.api.mock import (
    get_mock_agent_attributes,
    get_mock_agent_properties,
    get_mock_agent_state,
    get_mock_recurrent_state,
    get_mock_birdview,
    get_mock_infractions,
    get_mock_traffic_light_states,
    get_mock_light_recurrent_states
)
from invertedai.common import (
    AgentAttributes,
    AgentProperties,
    AgentState,
    Image,
    InfractionIndicators,
    LightRecurrentState,
    LightRecurrentStates,
    RecurrentState,
    TrafficLightStatesDict
)


class InitializeResponse(BaseModel):
    """
    Response returned from an API call to :func:`iai.initialize`.
    """
    agent_states: List[AgentState] #: Initial states of all initialized agents.
    recurrent_states: List[Optional[RecurrentState]] #: To pass to :func:`iai.drive` at the first time step.
    agent_attributes: List[Optional[AgentAttributes]] #: Static attributes of all initialized agents.
    agent_properties: List[AgentProperties]  #: Static agent properties of all initialized agents.
    birdview: Optional[Image] #: If `get_birdview` was set, this contains the resulting image.
    infractions: Optional[List[InfractionIndicators]] #: If `get_infractions` was set, they are returned here.
    traffic_lights_states: Optional[TrafficLightStatesDict] #: Traffic light states for the full map, each key-value pair corresponds to one particular traffic light.
    light_recurrent_states: Optional[LightRecurrentStates] #: Light recurrent states for the full map. Pass this to :func:`iai.drive` at the first time step to let the server generate a realistic continuation of the traffic light state sequence. This does not work correctly if any specific light states were specified as input to `initialize`.
    api_model_version: str #: Model version used for this API call


@validate_call
def initialize(
    location: str,
    agent_attributes: Optional[List[AgentAttributes]] = None,
    agent_properties: Optional[List[AgentProperties]] = None,
    states_history: Optional[List[List[AgentState]]] = None,
    traffic_light_state_history: Optional[List[TrafficLightStatesDict]] = None,
    get_birdview: bool = False,
    location_of_interest: Optional[Tuple[float, float]] = None,
    get_infractions: bool = False,
    agent_count: Optional[int] = None,
    random_seed: Optional[int] = None,
    api_model_version: Optional[str] = None  # Model version used for this API call
) -> InitializeResponse:
    """
    Initializes a simulation in a given location, using a combination of **user-defined** and **sampled** agents.
    The `agent_properties` parameter is used to determine the agents that are placed into the given `location`. 
    **User-defined** agents are placed into a scene first, after which a number of agents are sampled conditionally.
    Any **user-defined** agent must have a corresponding fully specified static `AgentProperties` object in `agent_properties`.
    Furthermore for all **user-defined** agents, `states_history` must contain a list of `AgentState's` of all **user-defined** agents per historical time step. 
    Per desired **sampled** agent, an `AgentProperties` object must be provided at the end of `agent_properties` with only its `agent_type` specified.
    Agents are identified by their list index, so ensure the indices of each agent match in `states_history` and`agent_properties` when applicable. 
    The `agent_attributes` and `agent_count` parameters are deprecated.
    If traffic lights are present in the scene, their states history can be specified with a list of `TrafficLightStatesDict`, each 
    represent light states for one timestep, with the last element representing the current time step. It is legal to omit the traffic 
    light state specification, and the scene will be initialized with a light state configuration consistent with agent states.
    Every simulation must start with a call to this function in order to obtain correct recurrent states for :func:`drive`.

    Parameters
    ----------
    location:
        Location name in IAI format.

    agent_attributes:
        Deprecated. Static attributes for all agents.
        The pre-defined agents should be specified first, followed by the sampled agents.
        The optional waypoint passed will be ignored for Initialize.
    agent_properties:
        Agent properties for all agents, replacing the deprecated `agent_attributes`.
        The pre-defined agents should be specified first, followed by the sampled agents.
        The optional waypoint passed will be ignored for Initialize.
        max_speed: optional [float], the desired maximum speed of the agent in m/s.

    states_history:
        History of pre-defined agent states - the outer list is over time and the inner over agents,
        in chronological order, i.e., index 0 is the oldest state and index -1 is the current state.
        The order of agents should be the same as in `agent_attributes`.
        For best results, provide at least 10 historical states for each agent.

    traffic_light_state_history:
       History of traffic light states - the list is over time, in chronological order, i.e.
       the last element is the current state. If there are traffic lights in the map, 
       not specifying traffic light state is equivalent to using server generated light states.

    location_of_interest:
        Optional coordinates for spawning agents with the given location as center instead of the default map center

    get_birdview:
        If True, a birdview image will be returned representing the current world. Note this will significantly
        impact on the latency.

    get_infractions:
        If True, infraction metrics will be returned for each agent.

    agent_count:
        Deprecated. Equivalent to padding the `agent_attributes` list to this length with default `AgentAttributes`.

    random_seed:
        Controls the stochastic aspects of initialization for reproducibility.

    api_model_version:
        Optionally specify the version of the model. If None is passed which is by default, the best model will be used.

    See Also
    --------
    :func:`drive`
    :func:`location_info`
    :func:`light`
    :func:`blame`
    """

    if should_use_mock_api():
        assert agent_properties is not None or agent_attributes is not None or agent_count is not None
        if agent_properties is not None:
            agent_count = len(agent_properties)
        elif agent_attributes is not None:
            agent_count = len(agent_attributes)

        agent_properties = [get_mock_agent_properties() for _ in range(agent_count)]
        agent_attributes = [get_mock_agent_attributes() for _ in range(agent_count)]
        if agent_attributes is None:
            agent_states = [get_mock_agent_state() for _ in range(agent_count)]
        else:
            agent_states = states_history[-1] if states_history is not None else []
        recurrent_states = [get_mock_recurrent_state() for _ in range(agent_count)]
        birdview = get_mock_birdview()
        infractions = get_mock_infractions(len(agent_states))
        response = InitializeResponse(
            agent_states=agent_states,
            agent_attributes=agent_attributes,
            agent_properties=agent_properties,
            recurrent_states=recurrent_states,
            birdview=birdview,
            infractions=infractions,
            api_model_version=api_model_version if api_model_version is not None else "best",
            traffic_lights_states=traffic_light_state_history[-1] if traffic_light_state_history is not None else None,
            light_recurrent_states=get_mock_light_recurrent_states(len(traffic_light_state_history[0])) if traffic_light_state_history is not None else None
        )
        return response

    if agent_attributes is not None:
        warnings.warn('agent_attributes is deprecated. Please use agent_properties.',category=DeprecationWarning)

    model_inputs = dict(
        location=location,
        num_agents_to_spawn=agent_count,
        states_history=states_history
        if states_history is None
        else [[st.tolist() for st in states] for states in states_history],
        agent_attributes=agent_attributes
        if agent_attributes is None
        else [state.tolist() for state in agent_attributes],
        agent_properties=agent_properties if agent_properties is None 
        else [ap.serialize() if ap else None for ap in agent_properties] ,
        traffic_light_state_history=traffic_light_state_history,
        get_birdview=get_birdview,
        location_of_interest=location_of_interest,
        get_infractions=get_infractions,
        random_seed=random_seed,
        model_version=api_model_version if api_model_version is not None else "best"
    )
    start = time.time()
    timeout = TIMEOUT
    while True:
        try:
            response = iai.session.request(model="initialize", data=model_inputs)
            response = InitializeResponse(
                agent_states=[
                    AgentState.fromlist(state) for state in response["agent_states"]
                ],
                agent_attributes=[
                    AgentAttributes.fromlist(attr) for attr in response["agent_attributes"]
                ] if response["agent_attributes"] is not None else [],
                agent_properties=[
                    AgentProperties.deserialize(ap) for ap in response["agent_properties"]
                ],
                recurrent_states=[
                    RecurrentState.fromval(r) for r in response["recurrent_states"]
                ],
                birdview=Image.fromval(response["birdview"])
                if response["birdview"] is not None
                else None,
                infractions=[
                    InfractionIndicators.fromlist(infractions)
                    for infractions in response["infraction_indicators"]
                ]
                if response["infraction_indicators"]
                else [],
                api_model_version=response["model_version"],
                traffic_lights_states=response["traffic_lights_states"] 
                if response["traffic_lights_states"] is not None 
                else None,
                light_recurrent_states=[
                    LightRecurrentState(state=state_arr[0], time_remaining=state_arr[1]) 
                    for state_arr in response["light_recurrent_states"]
                ] 
                if response["light_recurrent_states"] is not None 
                else None
            )
            return response
        except TryAgain as e:
            if timeout is not None and time.time() > start + timeout:
                raise e
            iai.logger.info(iai.logger.logfmt("Waiting for model to warm up", error=e))


@validate_call
async def async_initialize(
    location: str,
    agent_attributes: Optional[List[AgentAttributes]] = None,
    agent_properties: Optional[List[AgentProperties]] = None,
    states_history: Optional[List[List[AgentState]]] = None,
    traffic_light_state_history: Optional[List[TrafficLightStatesDict]] = None,
    get_birdview: bool = False,
    location_of_interest: Optional[Tuple[float, float]] = None,
    get_infractions: bool = False,
    agent_count: Optional[int] = None,
    random_seed: Optional[int] = None,
    api_model_version: Optional[str] = None
) -> InitializeResponse:
    """
    The async version of :func:`initialize`
    """

    model_inputs = dict(
        location=location,
        num_agents_to_spawn=agent_count,
        states_history=states_history
        if states_history is None
        else [[st.tolist() for st in states] for states in states_history],
        agent_attributes=agent_attributes
        if agent_attributes is None
        else [state.tolist() for state in agent_attributes],
        agent_properties=agent_properties if agent_properties is None 
        else [ap.serialize() if ap else None for ap in agent_properties] ,
        traffic_light_state_history=traffic_light_state_history,
        get_birdview=get_birdview,
        location_of_interest=location_of_interest,
        get_infractions=get_infractions,
        random_seed=random_seed,
        model_version=api_model_version
    )

    response = await iai.session.async_request(model="initialize", data=model_inputs)
    agents_spawned = len(response["agent_states"])
    if agents_spawned != agent_count:
        iai.logger.warning(
            f"Unable to spawn a scenario for {agent_count} agents,  {agents_spawned} spawned instead."
        )
    response = InitializeResponse(
        agent_states=[
            AgentState.fromlist(state) for state in response["agent_states"]
        ],
        agent_attributes=[
            AgentAttributes.fromlist(attr) for attr in response["agent_attributes"]
        ],
        agent_properties=[
                    AgentProperties.deserialize(ap) for ap in response["agent_properties"]
                ],
        recurrent_states=[
            RecurrentState.fromval(r) for r in response["recurrent_states"]
        ],
        birdview=Image.fromval(response["birdview"])
        if response["birdview"] is not None
        else None,
        infractions=[
            InfractionIndicators.fromlist(infractions)
            for infractions in response["infraction_indicators"]
        ]
        if response["infraction_indicators"]
        else [],
        api_model_version=response["model_version"],
        traffic_lights_states=response["traffic_lights_states"]
        if response["traffic_lights_states"] is not None 
        else None,
        light_recurrent_states=[
            LightRecurrentState(state=state_arr[0], time_remaining=state_arr[1]) 
            for state_arr in response["light_recurrent_states"]
            ] 
        if response["light_recurrent_states"] is not None 
        else None
    )
    return response



import time
import asyncio
import warnings
from typing import List, Optional, Tuple
from pydantic import BaseModel, validate_call

import invertedai as iai
from invertedai.api.config import TIMEOUT, should_use_mock_api
from invertedai.error import APIConnectionError, InvalidInput
from invertedai.api.mock import (
    mock_update_agent_state,
    get_mock_birdview,
    get_mock_infractions,
    get_mock_light_recurrent_states
)
from invertedai.common import (
    AgentState,
    RecurrentState,
    Image,
    InfractionIndicators,
    AgentAttributes,
    AgentProperties,
    TrafficLightStatesDict,
    LightRecurrentStates,
    LightRecurrentState,
)


class DriveResponse(BaseModel):
    """
    Response returned from an API call to :func:`iai.drive`.
    """

    agent_states: List[AgentState] #: Predicted states for all agents at the next time step.
    recurrent_states: List[RecurrentState] #: To pass to :func:`iai.drive` at the subsequent time step.
    birdview: Optional[Image] #: If `get_birdview` was set, this contains the resulting image.
    infractions: Optional[List[InfractionIndicators]]  #: If `get_infractions` was set, they are returned here.
    is_inside_supported_area: List[bool] #: For each agent, indicates whether the predicted state is inside supported area.
    traffic_lights_states: Optional[TrafficLightStatesDict] #: Traffic light states for the full map, as seen by the agents before they performed their actions resulting in the returned state. Each key-value pair corresponds to one particular traffic light.
    light_recurrent_states: Optional[LightRecurrentStates] #: Light recurrent states for the full map, each element corresponds to one light group. Pass this to the next call of :func:`iai.drive` for the server to realistically update the traffic light states.
    api_model_version: str # Model version used for this API call


@validate_call
def drive(
    location: str,
    agent_states: List[AgentState],
    agent_attributes: Optional[List[AgentAttributes]] = None,
    agent_properties: Optional[List[AgentProperties]] = None,
    recurrent_states: Optional[List[RecurrentState]] = None,
    traffic_lights_states: Optional[TrafficLightStatesDict] = None,
    light_recurrent_states: Optional[LightRecurrentStates] = None,
    get_birdview: bool = False,
    rendering_center: Optional[Tuple[float, float]] = None,
    rendering_fov: Optional[float] = None,
    get_infractions: bool = False,
    random_seed: Optional[int] = None,
    api_model_version: Optional[str] = None
) -> DriveResponse:
    """
    Update the state of all given agents forward one time step. Agents are identified by their list index.

    Parameters
    ----------
    location:
        Location name in IAI format.

    agent_states:
        Current states of all agents.
        The state must include x: [float], y: [float] coordinate in meters
        orientation: [float] in radians with 0 pointing along x and pi/2 pointing along y and
        speed: [float] in m/s.

    agent_attributes:
        Deprecated. Static attributes of all agents.
        List of agent attributes. Each agent requires, length: [float]
        width: [float] and rear_axis_offset: [float] all in meters. agent_type: [str],
        currently supports 'car' and 'pedestrian'.
        waypoint: optional [Point], the target waypoint of the agent.

    agent_properties:
        Agent properties for all agents, replacing soon to be deprecated `agent_attributes`.
        List of agent attributes. Each agent requires, length: [float]
        width: [float] and rear_axis_offset: [float] all in meters. agent_type: [str],
        currently supports 'car' and 'pedestrian'.
        waypoint: optional [Point], the target waypoint of the agent.
        max_speed: optional [float], the desired maximum speed of the agent in m/s.

    recurrent_states:
        Recurrent states for all agents, obtained from the previous call to
        :func:`drive` or :func:`initialize`.

    get_birdview:
        Whether to return an image visualizing the simulation state.
        This is very slow and should only be used for debugging.

    rendering_center:
        Optional center coordinates for the rendered birdview.

    rendering_fov:
        Optional fov for the rendered birdview.

    get_infractions:
        Whether to check predicted agent states for infractions.
        This introduces some overhead, but it should be relatively small.

    traffic_lights_states:
       If the location contains traffic lights within the supported area,
       their current state can be provided here. It is legal to not provide this field, and use
       'light_recurrent_states' to step the traffic lights. If provided, light states from 'traffic_light_states' will override
       the original light states given by 'light_recurrent_states'. The server does not currently support continuing user-provided light state sequences, 
       so once the states are provided at any step, they should also be provided on all subsequent steps to guarantee coherent light sequences.
       If neither 'traffic_lights_states' nor 'light_recurrent_states' are provided, the server will arbitrarily initialize the traffic light states,
       and return the associated 'light_recurrent_states' in the response.

    light_recurrent_states:
       Light recurrent states for all agents, obtained from the previous call to
        :func:`drive` or :func:`initialize`.
       Specifies the state and time remaining for each light group in the map.
       To let the server manage all light states in the scene, 
       pass 'light_recurrent_states' from the previous response of :func:`drive` here and leave `traffic_light_states=None`.

    random_seed:
        Controls the stochastic aspects of agent behavior for reproducibility.

    api_model_version:
        Optionally specify the version of the model. If None is passed which is by default, the best model will be used.
    See Also
    --------
    :func:`initialize`
    :func:`location_info`
    :func:`light`
    :func:`blame`
    """

    if should_use_mock_api():
        agent_states = [mock_update_agent_state(s) for s in agent_states]
        present_mask = [True for _ in agent_states]
        birdview = get_mock_birdview()
        infractions = get_mock_infractions(len(agent_states))
        response = DriveResponse(
            agent_states=agent_states,
            is_inside_supported_area=present_mask,
            recurrent_states=recurrent_states,
            birdview=birdview,
            infractions=infractions,
            traffic_lights_states=traffic_lights_states if traffic_lights_states is not None else None,
            light_recurrent_states=get_mock_light_recurrent_states(len(traffic_lights_states)) if traffic_lights_states is not None else None,
            api_model_version=api_model_version if api_model_version is not None else "best"
        )
        return response

    if agent_attributes is not None:
        warnings.warn('agent_attributes is deprecated. Please use agent_properties.',category=DeprecationWarning) 

    def _tolist(input_data: List):
        if not isinstance(input_data, list):
            return input_data.tolist()
        else:
            return input_data

    recurrent_states = _tolist(recurrent_states) if recurrent_states is not None else None
    model_inputs = dict(
        location=location,
        agent_states=[state.tolist() for state in agent_states],
        agent_attributes=[state.tolist() for state in agent_attributes] if agent_attributes is not None else None,
        agent_properties=[ap.serialize() for ap in agent_properties] if agent_properties is not None else None,
        recurrent_states=[r.packed for r in recurrent_states] if recurrent_states is not None else None,
        traffic_lights_states=traffic_lights_states,
        light_recurrent_states=[light_recurrent_state.tolist() for light_recurrent_state in light_recurrent_states] 
        if light_recurrent_states is not None else None,
        get_birdview=get_birdview,
        get_infractions=get_infractions,
        random_seed=random_seed,
        rendering_center=rendering_center,
        rendering_fov=rendering_fov,
        model_version=api_model_version
    )
    start = time.time()
    timeout = TIMEOUT

    while True:
        try:
            response = iai.session.request(model="drive", data=model_inputs)

            response = DriveResponse(
                agent_states=[
                    AgentState.fromlist(state) for state in response["agent_states"]
                ],
                recurrent_states=[
                    RecurrentState.fromval(r) for r in response["recurrent_states"]
                ],
                birdview=Image.fromval(response["birdview"])
                if response["birdview"] is not None
                else None,
                infractions=[
                    InfractionIndicators.fromlist(infractions)
                    for infractions in response["infraction_indicators"]
                ]
                if response["infraction_indicators"]
                else [],
                is_inside_supported_area=response["is_inside_supported_area"],
                api_model_version=response["model_version"],
                traffic_lights_states=response["traffic_lights_states"]
                if response["traffic_lights_states"] is not None 
                else None,
                light_recurrent_states=[
                    LightRecurrentState(state=state_arr[0], time_remaining=state_arr[1]) 
                    for state_arr in response["light_recurrent_states"]
                ] 
                if response["light_recurrent_states"] is not None 
                else None
            )

            return response
        except APIConnectionError as e:
            iai.logger.warning("Retrying")
            if (
                    timeout is not None and time.time() > start + timeout
            ) or not e.should_retry:
                raise e


@validate_call
async def async_drive(
    location: str,
    agent_states: List[AgentState],
    agent_attributes: Optional[List[AgentAttributes]]=None,
    agent_properties: Optional[List[AgentProperties]]=None,
    recurrent_states: Optional[List[RecurrentState]] = None,
    traffic_lights_states: Optional[TrafficLightStatesDict] = None,
    light_recurrent_states: Optional[LightRecurrentStates] = None,
    get_birdview: bool = False,
    rendering_center: Optional[Tuple[float, float]] = None,
    rendering_fov: Optional[float] = None,
    get_infractions: bool = False,
    random_seed: Optional[int] = None,
    api_model_version: Optional[str] = None
) -> DriveResponse:
    """
    A light async version of :func:`drive`
    """

    def _tolist(input_data: List):
        if not isinstance(input_data, list):
            return input_data.tolist()
        else:
            return input_data

    recurrent_states = _tolist(recurrent_states) if recurrent_states is not None else None
    model_inputs = dict(
        location=location,
        agent_states=[state.tolist() for state in agent_states],
        agent_attributes=[state.tolist() for state in agent_attributes] if agent_attributes is not None else None,
        agent_properties=[ap.serialize() for ap in agent_properties] if agent_properties is not None else None,
        recurrent_states=[r.packed for r in recurrent_states] if recurrent_states is not None else None,
        traffic_lights_states=traffic_lights_states,
        light_recurrent_states=[light_recurrent_state.tolist() for light_recurrent_state in light_recurrent_states] 
        if light_recurrent_states is not None else None,
        get_birdview=get_birdview,
        get_infractions=get_infractions,
        random_seed=random_seed,
        rendering_center=rendering_center,
        rendering_fov=rendering_fov,
        model_version=api_model_version
    )
    response = await iai.session.async_request(model="drive", data=model_inputs)

    response = DriveResponse(
        agent_states=[
            AgentState.fromlist(state) for state in response["agent_states"]
        ],
        recurrent_states=[
            RecurrentState.fromval(r) for r in response["recurrent_states"]
        ],
        birdview=Image.fromval(response["birdview"])
        if response["birdview"] is not None
        else None,
        infractions=[
            InfractionIndicators.fromlist(infractions)
            for infractions in response["infraction_indicators"]
        ]
        if response["infraction_indicators"]
        else [],
        is_inside_supported_area=response["is_inside_supported_area"],
        api_model_version=response["model_version"],
        traffic_lights_states=response["traffic_lights_states"] 
        if response["traffic_lights_states"] is not None 
        else None,
        light_recurrent_states=[
            LightRecurrentState(state=state_arr[0], time_remaining=state_arr[1]) 
            for state_arr in response["light_recurrent_states"]
        ] 
        if response["light_recurrent_states"] is not None 
        else None
    )

    return response



import os
import warnings
import importlib.metadata
__version__ = importlib.metadata.version("invertedai")
from distutils.util import strtobool

from invertedai.api.light import light
from invertedai.api.location import location_info
from invertedai.api.initialize import initialize, async_initialize
from invertedai.api.drive import drive, async_drive
from invertedai.api.blame import blame, async_blame
from invertedai.cosimulation import BasicCosimulation
from invertedai.utils import Jupyter_Render, IAILogger, Session
from invertedai.large.initialize import (
    get_regions_in_grid, 
    get_number_of_agents_per_region_by_drivable_area, 
    get_regions_default, 
    large_initialize
)
from invertedai.large.drive import large_drive
from invertedai.logs.logger import LogWriter, LogReader
from invertedai.logs.debug_logger import DebugLogger

warnings.filterwarnings(action="once",message=".*agent_attributes.*")

dev = strtobool(os.environ.get("IAI_DEV", "false"))
if dev:
    dev_url = os.environ.get("IAI_DEV_URL", "http://localhost:8000")
commercial_url = "https://api.inverted.ai/v0/aws/m1"
academic_url = "https://api.inverted.ai/v0/academic/m1"

log_level = os.environ.get("IAI_LOG_LEVEL", "WARNING")
log_console = strtobool(os.environ.get("IAI_LOG_CONSOLE", "true"))
log_file = strtobool(os.environ.get("IAI_LOG_FILE", "false"))
api_key = os.environ.get("IAI_API_KEY", "")
debug_logger_path = os.environ.get("IAI_LOGGER_PATH", None)

debug_logger = None
if debug_logger_path is not None:
    debug_logger = DebugLogger(debug_logger_path)
logger = IAILogger(level=log_level, consoel=bool(log_console), log_file=bool(log_file))

session = Session(debug_logger)
if api_key:
    session.add_apikey(api_key)
add_apikey = session.add_apikey
use_mock_api = session.use_mock_api

if strtobool(os.environ.get("IAI_MOCK_API", "false")):
    use_mock_api()

model_resources = {
    "initialize": ("post", "/initialize"),
    "blame": ("post", "/blame"),
    "drive": ("post", "/drive"),
    "location_info": ("get", "/location_info"),
    "light": ("get", "/light"),
    "test": ("get", "/test"),
}
__all__ = [
    "BasicCosimulation",
    "Jupyter_Render",
    "logger",
    "session",
    "add_apikey",
    "use_mock_api",
    "blame",
    "drive",
    "initialize",
    "location_info",
    "light",
    "async_initialize",
    "async_drive",
    "async_blame",
]